import unittest
from secator.celery import app, forward_results
from secator.utils_test import mock_command, FIXTURES_TASKS, TEST_TASKS, FIXTURES_DIR, load_fixture
from secator.output_types import Url
from celery import chain, chord

TARGETS = ['bing.com', 'google.com', 'wikipedia.org', 'ibm.com', 'cnn.com', 'karate.com']


class TestCelery(unittest.TestCase):

	def test_httpx_chain(self):
		from secator.tasks import httpx
		if httpx not in TEST_TASKS:
			return

		with mock_command(httpx, fixture=[FIXTURES_TASKS[httpx]] * len(TARGETS)):
			sigs = [forward_results.si([])] + [httpx.s(target) for target in TARGETS]
			workflow = chain(*sigs)
			result = workflow.apply()
			results = result.get()
			urls = [r.url for r in results if r._type == 'url']
			targets = [r.name for r in results if r._type == 'target']
			self.assertEqual(len(urls), len(TARGETS))
			self.assertEqual(len(targets), len(TARGETS))

	def test_httpx_chain_with_results(self):
		from secator.tasks import httpx
		if httpx not in TEST_TASKS:
			return

		existing_results = [Url(**{
			"url": "https://example.synology.me",
			"method": "GET",
			"status_code": 200,
			"words": 438,
			"lines": 136,
			"content_type":
			"text/html",
			"content_length": 11577,
			"host": "82.66.157.114",
			"time": 0.16246860100000002,
			"_source": "httpx",
			"_type": "url"
		})]
		with mock_command(httpx, fixture=[FIXTURES_TASKS[httpx]] * len(TARGETS)):
			sigs = [forward_results.s(existing_results)] + [httpx.s(target) for target in TARGETS]
			workflow = chain(*sigs)
			result = workflow.apply()
			results = result.get()
			urls = [r.url for r in results if r._type == 'url']
			targets = [r.name for r in results if r._type == 'target']
			self.assertEqual(len(urls), len(TARGETS) + 1)
			self.assertEqual(len(targets), len(TARGETS))
			self.assertIn(existing_results[0], results)

	def test_httpx_workflow(self):
		from secator.tasks import httpx
		if httpx not in TEST_TASKS:
			return

		targets = ['bing.com', 'google.com', 'wikipedia.org', 'ibm.com', 'cnn.com', 'karate.com']
		sigs = []
		for target in targets:
			sig = httpx().s(target)
			sigs.append(sig)
		with mock_command(httpx, fixture=[FIXTURES_TASKS[httpx]] * len(targets)):
			workflow = chain(
				forward_results.s([]),
				sigs[0],
				chord((
					sigs[1],
					sigs[2],
				), forward_results.s()),
				sigs[3],
				chord((
					sigs[4],
					sigs[5],
				), forward_results.s())
			)
			result = workflow.apply()
			results = result.get()
			urls = [r.url for r in results if r._type == 'url']
			targets = [r.name for r in results if r._type == 'target']
			self.assertEqual(len(targets), len(TARGETS))
			self.assertEqual(len(urls), len(TARGETS))

	def test_nmap_workflow(self):
		from secator.tasks import nmap
		if nmap not in TEST_TASKS:
			return

		nmap_fixture = load_fixture('nmap_output', fixtures_dir=FIXTURES_DIR, ext='.xml', only_path=True)
		with mock_command(nmap, fixture=[]):
			workflow = chain(
				forward_results.s([]),
				chord((
					nmap.s(TARGETS, output_path=nmap_fixture)
				), forward_results.s()),
			)
			result = workflow.apply()
			results = result.get()
			targets = [r.name for r in results if r._type == 'target']
			self.assertEqual(len(targets), len(TARGETS))

	def test_ffuf_chunked(self):
		from secator.tasks import ffuf
		with mock_command(ffuf, fixture=[]):
			workflow = chain(
				forward_results.s([]),
				chord((
					ffuf.s(TARGETS)
				), forward_results.s()),
			)
			result = workflow.apply()
			results = result.get()
			self.assertEqual(len(results), len(TARGETS))



from dataclasses import dataclass, field
import time
from secator.output_types import OutputType
from secator.utils import rich_to_ansi, traceback_as_string, rich_escape as _s


@dataclass
class Error(OutputType):
	message: str
	traceback: str = field(default='', compare=False)
	traceback_title: str = field(default='', compare=False)
	_source: str = field(default='', repr=True)
	_type: str = field(default='error', repr=True)
	_timestamp: int = field(default_factory=lambda: time.time(), compare=False)
	_uuid: str = field(default='', repr=True, compare=False)
	_context: dict = field(default_factory=dict, repr=True, compare=False)
	_duplicate: bool = field(default=False, repr=True, compare=False)
	_related: list = field(default_factory=list, compare=False)

	_table_fields = ['message', 'traceback']
	_sort_by = ('_timestamp',)

	def from_exception(e, **kwargs):
		errtype = type(e).__name__
		message = errtype
		if str(e):
			message += f': {str(e)}'
		traceback = traceback_as_string(e) if errtype not in ['KeyboardInterrupt', 'GreenletExit'] else ''
		error = Error(message=message, traceback=traceback, **kwargs)
		return error

	def __str__(self):
		return self.message

	def __repr__(self):
		s = rf"\[[bold red]ERR[/]] {_s(self.message)}"
		if self.traceback:
			s += ':'
			traceback_pretty = '   ' + _s(self.traceback).replace('\n', '\n   ')
			if self.traceback_title:
				traceback_pretty = f'   {self.traceback_title}:\n{traceback_pretty}'
			s += f'\n[dim]{_s(traceback_pretty)}[/]'
		return rich_to_ansi(s)



import unittest
from secator.output_types import Url, Target, Port, Vulnerability, Info
from secator.runners import Command
from secator.serializers import JSONSerializer
from time import sleep
from threading import Thread
import queue

class TestWorker(unittest.TestCase):

	@classmethod
	def setUpClass(cls):
		cls.queue = queue.Queue()
		cls.cmd = Command.execute('secator worker', name='secator_worker', quiet=True, run=False)
		cls.thread = Thread(target=cls.cmd.run)
		cls.thread.start()
		sleep(3)

	@classmethod
	def tearDownClass(cls) -> None:
		cls.cmd.stop_process()
		cls.thread.join()

	def test_httpx(self):
		cmd = Command.execute(
			'secator x httpx testphp.vulnweb.com -json',
			name='secator_x_httpx',
			no_process=False,
			quiet=True,
			cls_attributes={'output_types': [Target, Url, Info], 'item_loaders': [JSONSerializer()]}
		)
		# self.assertEqual(cmd.return_code, 0)  # TODO: figure out why return code is -9 when running from unittest
		self.assertEqual(len(cmd.findings), 1)
		url = Url(
			'http://testphp.vulnweb.com',
			status_code=200,
			title='Home of Acunetix Art',
			webserver='nginx',
			tech=['DreamWeaver', 'Nginx:1.19.0', 'PHP:5.6.40', 'Ubuntu'],
			content_type='text/html',
			content_length=4958,
			_source='httpx'
		)
		self.assertIn(url, cmd.results)

	def test_host_recon(self):
		cmd = Command.execute(
			'secator w host_recon vulnweb.com -json -p 80 -tid nginx-version',
			name='secator_w_host_recon',
			no_process=False,
			quiet=True,
			cls_attributes={'output_types': [Target, Url, Port, Vulnerability, Info], 'item_loaders': [JSONSerializer()]}
		)
		# self.assertEqual(cmd.return_code, 0)  # TODO: ditto
		self.assertGreater(len(cmd.results), 0)
		port = Port(
			port=80,
			ip="44.228.249.3",
			state="open",
			service_name="nginx/1.19.0",
			_source="nmap"
		)
		url = Url(
			'http://vulnweb.com',
			status_code=200,
			title='Acunetix Web Vulnerability Scanner - Test Websites',
			webserver='nginx/1.19.0',
			tech=['Nginx:1.19.0'],
			content_type='text/html',
			content_length=4018,
			_source='httpx'
		)
		vuln = Vulnerability(
			name='nginx-version',
			provider='',
			id='',
			matched_at='http://vulnweb.com',
			confidence='high',
			confidence_nb=4,
			severity_nb=4,
			severity='info',
			tags=['tech', 'nginx'],
			_source='nuclei_url'
		)
		self.assertIn(port, cmd.results)
		self.assertIn(url, cmd.results)
		self.assertIn(vuln, cmd.results)

	# def test_pd_pipe(self):
	# 	cmd = Command.execute(
	# 		'secator x subfinder vulnweb.com | secator x nmap | secator x httpx | secator x katana | secator x httpx | secator x gf --pattern lfi -fmt "{match}" | secator x dalfox'
	# 	)



import unittest
from secator.serializers.dataclass import dumps_dataclass, loads_dataclass
from secator.output_types import Port, Vulnerability


class TestSerializers(unittest.TestCase):

	def test_dumps_loads(self):
		results = [
			Port(port=53, ip='127.0.0.1', host='localhost'),
			Vulnerability(matched_at='localhost', name='CVE-123123123', provider='nmap')
		]
		results = dumps_dataclass(results)
		results = loads_dataclass(results)
		self.assertTrue(isinstance(results, list))
		self.assertEqual(len(results), 2)
		self.assertTrue(all(type(result) in [Port, Vulnerability] for result in results))
		self.assertTrue(isinstance(results, list))

	def test_dumps_loads_nested(self):
		results = {
			'info': {'name': 'test'},
			'results': {
				'ports': [
					{'port': 53, 'ip': '127.0.0.1', 'host': 'localhost', '_type': 'port'},
				],
				'vulnerabilities': [
					{'matched_at': 'localhost', 'name': 'CVE-123123123', 'provider': 'nmap', '_type': 'vulnerability'}
				]
			}
		}
		results = loads_dataclass(dumps_dataclass(results))
		self.assertTrue(isinstance(results['results']['ports'][0], Port))
		self.assertTrue(isinstance(results['results']['vulnerabilities'][0], Vulnerability))

	def test_dumps_loads_nested_obj(self):
		results = {
			'info': {'name': 'test'},
			'results': {
				'ports': [
					Port(port=53, ip='127.0.0.1', host='localhost'),
				],
				'vulnerabilities': [
					Vulnerability(matched_at='localhost', name='CVE-123123123', provider='nmap')
				]
			}
		}
		results = loads_dataclass(dumps_dataclass(results))
		self.assertTrue(isinstance(results['results']['ports'][0], Port))
		self.assertTrue(isinstance(results['results']['vulnerabilities'][0], Vulnerability))


import copy
import getpass
import logging
import os
import re
import shlex
import signal
import subprocess
import sys
import uuid

from time import time

import psutil
from fp.fp import FreeProxy

from secator.definitions import OPT_NOT_SUPPORTED, OPT_PIPE_INPUT
from secator.config import CONFIG
from secator.output_types import Info, Warning, Error, Target, Stat
from secator.runners import Runner
from secator.template import TemplateLoader
from secator.utils import debug, rich_escape as _s


logger = logging.getLogger(__name__)


class Command(Runner):
	"""Base class to execute an external command."""
	# Base cmd
	cmd = None

	# Meta options
	meta_opts = {}

	# Additional command options
	opts = {}

	# Option prefix char
	opt_prefix = '-'

	# Option key map to transform option names
	opt_key_map = {}

	# Option value map to transform option values
	opt_value_map = {}

	# Output map to transform JSON output keys
	output_map = {}

	# Run in shell if True (not recommended)
	shell = False

	# Current working directory
	cwd = None

	# Output encoding
	encoding = 'utf-8'

	# Flag to take the input
	input_flag = None

	# Input path (if a file is constructed)
	input_path = None

	# Input chunk size
	input_chunk_size = CONFIG.runners.input_chunk_size

	# Input required
	input_required = True

	# Flag to take a file as input
	file_flag = None

	# Flag to enable output JSON
	json_flag = None

	# Flag to show version
	version_flag = None

	# Install
	install_pre = None
	install_post = None
	install_cmd = None
	install_github_handle = None

	# Serializer
	item_loader = None
	item_loaders = []

	# Hooks
	hooks = [
		'on_cmd',
		'on_cmd_done',
		'on_line'
	]

	# Ignore return code
	ignore_return_code = False

	# Return code
	return_code = -1

	# Output
	output = ''

	# Proxy options
	proxychains = False
	proxy_socks5 = False
	proxy_http = False

	# Profile
	profile = 'io'

	def __init__(self, inputs=[], **run_opts):

		# Build runnerconfig on-the-fly
		config = TemplateLoader(input={
			'name': self.__class__.__name__,
			'type': 'task',
			'description': run_opts.get('description', None)
		})

		# Extract run opts
		hooks = run_opts.pop('hooks', {})
		caller = run_opts.get('caller', None)
		results = run_opts.pop('results', [])
		context = run_opts.pop('context', {})
		self.skip_if_no_inputs = run_opts.pop('skip_if_no_inputs', False)

		# Prepare validators
		input_validators = []
		if not self.skip_if_no_inputs:
			input_validators.append(self._validate_input_nonempty)
		if not caller:
			input_validators.append(self._validate_chunked_input)
		validators = {'validate_input': input_validators}

		# Call super().__init__
		super().__init__(
			config=config,
			inputs=inputs,
			results=results,
			run_opts=run_opts,
			hooks=hooks,
			validators=validators,
			context=context)

		# Cmd name
		self.cmd_name = self.__class__.cmd.split(' ')[0]

		# Inputs path
		self.inputs_path = None

		# Current working directory for cmd
		self.cwd = self.run_opts.get('cwd', None)

		# Print cmd
		self.print_cmd = self.run_opts.get('print_cmd', False)

		# Stat update
		self.last_updated_stat = None

		# Process
		self.process = None

		# Proxy config (global)
		self.proxy = self.run_opts.pop('proxy', False)
		self.configure_proxy()

		# Build command input
		self._build_cmd_input()

		# Build command
		self._build_cmd()

		# Run on_cmd hook
		self.run_hooks('on_cmd')

		# Build item loaders
		instance_func = getattr(self, 'item_loader', None)
		item_loaders = self.item_loaders.copy()
		if instance_func:
			item_loaders.append(instance_func)
		self.item_loaders = item_loaders

	def toDict(self):
		res = super().toDict()
		res.update({
			'cmd': self.cmd,
			'cwd': self.cwd,
			'return_code': self.return_code
		})
		return res

	@classmethod
	def needs_chunking(cls, targets, sync):
		many_targets = len(targets) > 1
		targets_over_chunk_size = cls.input_chunk_size and len(targets) > cls.input_chunk_size
		has_file_flag = cls.file_flag is not None
		chunk_it = (sync and many_targets and not has_file_flag) or (not sync and many_targets and targets_over_chunk_size)
		return chunk_it

	@classmethod
	def delay(cls, *args, **kwargs):
		# TODO: Move this to TaskBase
		from secator.celery import run_command
		results = kwargs.get('results', [])
		kwargs['sync'] = False
		name = cls.__name__
		return run_command.apply_async(args=[results, name] + list(args), kwargs={'opts': kwargs}, queue=cls.profile)

	@classmethod
	def s(cls, *args, **kwargs):
		# TODO: Move this to TaskBase
		from secator.celery import run_command
		return run_command.s(cls.__name__, *args, opts=kwargs).set(queue=cls.profile)

	@classmethod
	def si(cls, *args, results=[], **kwargs):
		# TODO: Move this to TaskBase
		from secator.celery import run_command
		return run_command.si(results, cls.__name__, *args, opts=kwargs).set(queue=cls.profile)

	def get_opt_value(self, opt_name):
		return Command._get_opt_value(
			self.run_opts,
			opt_name,
			dict(self.opts, **self.meta_opts),
			opt_prefix=self.config.name)

	@classmethod
	def get_supported_opts(cls):
		def convert(d):
			for k, v in d.items():
				if hasattr(v, '__name__') and v.__name__ in ['str', 'int', 'float']:
					d[k] = v.__name__
			return d

		cls_opts = copy.deepcopy(cls.opts)
		opts = {k: convert(v) for k, v in cls_opts.items()}
		for k, v in opts.items():
			v['meta'] = cls.__name__
			v['supported'] = True

		cls_meta_opts = copy.deepcopy(cls.meta_opts)
		meta_opts = {k: convert(v) for k, v in cls_meta_opts.items() if cls.opt_key_map.get(k) is not OPT_NOT_SUPPORTED}
		for k, v in meta_opts.items():
			v['meta'] = 'meta'
			if cls.opt_key_map.get(k) is OPT_NOT_SUPPORTED:
				v['supported'] = False
			else:
				v['supported'] = True
		opts = dict(opts)
		opts.update(meta_opts)
		return opts

	#---------------#
	# Class methods #
	#---------------#

	@classmethod
	def execute(cls, cmd, name=None, cls_attributes={}, run=True, **kwargs):
		"""Execute an ad-hoc command.

		Can be used without defining an inherited class to run a command, while still enjoying all the good stuff in
		this class.

		Args:
			cls (object): Class.
			cmd (str): Command.
			name (str): Printed name.
			cls_attributes (dict): Class attributes.
			kwargs (dict): Options.

		Returns:
			secator.runners.Command: instance of the Command.
		"""
		name = name or cmd.split(' ')[0]
		kwargs['print_cmd'] = not kwargs.get('quiet', False)
		kwargs['print_line'] = True
		kwargs['no_process'] = kwargs.get('no_process', True)
		cmd_instance = type(name, (Command,), {'cmd': cmd, 'input_required': False})(**kwargs)
		for k, v in cls_attributes.items():
			setattr(cmd_instance, k, v)
		if run:
			cmd_instance.run()
		return cmd_instance

	def configure_proxy(self):
		"""Configure proxy. Start with global settings like 'proxychains' or 'random', or fallback to tool-specific
		proxy settings.

		TODO: Move this to a subclass of Command, or to a configurable attribute to pass to derived classes as it's not
		related to core functionality.
		"""
		opt_key_map = self.opt_key_map
		proxy_opt = opt_key_map.get('proxy', False)
		support_proxy_opt = proxy_opt and proxy_opt != OPT_NOT_SUPPORTED
		proxychains_flavor = getattr(self, 'proxychains_flavor', CONFIG.http.proxychains_command)
		proxy = False

		if self.proxy in ['auto', 'proxychains'] and self.proxychains:
			self.cmd = f'{proxychains_flavor} {self.cmd}'
			proxy = 'proxychains'

		elif self.proxy and support_proxy_opt:
			if self.proxy in ['auto', 'socks5'] and self.proxy_socks5 and CONFIG.http.socks5_proxy:
				proxy = CONFIG.http.socks5_proxy
			elif self.proxy in ['auto', 'http'] and self.proxy_http and CONFIG.http.http_proxy:
				proxy = CONFIG.http.http_proxy
			elif self.proxy == 'random' and self.proxy_http:
				proxy = FreeProxy(timeout=CONFIG.http.freeproxy_timeout, rand=True, anonym=True).get()
			elif self.proxy.startswith(('http://', 'socks5://')):
				proxy = self.proxy

		if proxy != 'proxychains':
			self.run_opts['proxy'] = proxy

		if proxy != 'proxychains' and self.proxy and not proxy:
			self._print(
				f'[bold red]Ignoring proxy "{self.proxy}" for {self.cmd_name} (not supported).[/]', rich=True)

	#----------#
	# Internal #
	#----------#
	def yielder(self):
		"""Run command and yields its output in real-time. Also saves the command line, return code and output to the
		database.

		Args:
			cmd (str): Command to run.
			cwd (str, Optional): Working directory to run from.
			shell (bool, Optional): Run command in a shell.
			history_file (str): History file path.
			mapper_func (Callable, Optional): Function to map output before yielding.
			encoding (str, Optional): Output encoding.
			ctx (dict, Optional): Scan context.

		Yields:
			str: Command stdout / stderr.
			dict: Serialized object.
		"""
		try:

			# Abort if it has children tasks
			if self.has_children:
				return

			# Print task description
			self.print_description()

			# Abort if no inputs
			if len(self.inputs) == 0 and self.skip_if_no_inputs:
				yield Warning(message=f'{self.unique_name} skipped (no inputs)', _source=self.unique_name, _uuid=str(uuid.uuid4()))
				return

			# Yield targets
			for input in self.inputs:
				yield Target(name=input, _source=self.unique_name, _uuid=str(uuid.uuid4()))

			# Check for sudo requirements and prepare the password if needed
			sudo_password, error = self._prompt_sudo(self.cmd)
			if error:
				yield Error(
					message=error,
					_source=self.unique_name,
					_uuid=str(uuid.uuid4())
				)
				return

			# Prepare cmds
			command = self.cmd if self.shell else shlex.split(self.cmd)

			# Check command is installed and auto-install
			if not self.no_process and not self.is_installed():
				if CONFIG.security.auto_install_commands:
					from secator.installer import ToolInstaller
					yield Info(
						message=f'Command {self.name} is missing but auto-installing since security.autoinstall_commands is set',  # noqa: E501
						_source=self.unique_name,
						_uuid=str(uuid.uuid4())
					)
					status = ToolInstaller.install(self.__class__)
					if not status.is_ok():
						yield Error(
							message=f'Failed installing {self.cmd_name}',
							_source=self.unique_name,
							_uuid=str(uuid.uuid4())
						)
						return

			# Output and results
			self.return_code = 0
			self.killed = False

			# Run the command using subprocess
			env = os.environ
			self.process = subprocess.Popen(
				command,
				stdin=subprocess.PIPE if sudo_password else None,
				stdout=subprocess.PIPE,
				stderr=subprocess.STDOUT,
				universal_newlines=True,
				shell=self.shell,
				env=env,
				cwd=self.cwd)
			self.print_command()

			# If sudo password is provided, send it to stdin
			if sudo_password:
				self.process.stdin.write(f"{sudo_password}\n")
				self.process.stdin.flush()

			# Process the output in real-time
			for line in iter(lambda: self.process.stdout.readline(), b''):
				# sleep(0)  # for async to give up control
				if not line:
					break
				yield from self.process_line(line)

			# Run hooks after cmd has completed successfully
			result = self.run_hooks('on_cmd_done')
			if result:
				yield from result

		except FileNotFoundError as e:
			yield from self.handle_file_not_found(e)

		except BaseException as e:
			self.debug(f'{self.unique_name}: {type(e).__name__}.', sub='error')
			self.stop_process()
			yield Error.from_exception(e, _source=self.unique_name, _uuid=str(uuid.uuid4()))

		finally:
			yield from self._wait_for_end()

	def is_installed(self):
		"""Check if a command is installed by using `which`.

		Args:
			command (str): The command to check.

		Returns:
			bool: True if the command is installed, False otherwise.
		"""
		result = subprocess.Popen(["which", self.cmd_name], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		result.communicate()
		return result.returncode == 0

	def process_line(self, line):
		"""Process a single line of output emitted on stdout / stderr and yield results."""

		# Strip line endings
		line = line.rstrip()

		# Some commands output ANSI text, so we need to remove those ANSI chars
		if self.encoding == 'ansi':
			ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
			line = ansi_escape.sub('', line)
			line = line.replace('\\x0d\\x0a', '\n')

		# Run on_line hooks
		line = self.run_hooks('on_line', line)
		if line is None:
			return

		# Run item_loader to try parsing as dict
		item_count = 0
		for item in self.run_item_loaders(line):
			yield item
			item_count += 1

		# Yield line if no items were yielded
		if item_count == 0:
			yield line

		# Skip rest of iteration (no process mode)
		if self.no_process:
			return

		# Yield command stats (CPU, memory, conns ...)
		# TODO: enable stats support with timer
		if self.last_updated_stat and (time() - self.last_updated_stat) < CONFIG.runners.stat_update_frequency:
			return

		yield from self.stats()
		self.last_updated_stat = time()

	def print_description(self):
		"""Print description"""
		if self.sync and not self.has_children:
			if self.caller and self.description:
				self._print(f'\n[bold gold3]:wrench: {self.description} [dim cyan]({self.config.name})[/][/] ...', rich=True)

	def print_command(self):
		"""Print command."""
		if self.print_cmd:
			cmd_str = _s(self.cmd)
			if self.sync and self.chunk and self.chunk_count:
				cmd_str += f' [dim gray11]({self.chunk}/{self.chunk_count})[/]'
			self._print(cmd_str, color='bold cyan', rich=True)
		self.debug('Command', obj={'cmd': self.cmd}, sub='init')

	def handle_file_not_found(self, exc):
		"""Handle case where binary is not found.

		Args:
			exc (FileNotFoundError): the exception.

		Yields:
			secator.output_types.Error: the error.
		"""
		self.return_code = 127
		if self.config.name in str(exc):
			message = 'Executable not found.'
			if self.install_cmd:
				message += f' Install it with "secator install tools {self.config.name}".'
			error = Error(message=message)
		else:
			error = Error.from_exception(exc)
		error._source = self.unique_name
		error._uuid = str(uuid.uuid4())
		yield error

	def stop_process(self):
		"""Sends SIGINT to running process, if any."""
		if not self.process:
			return
		self.debug(f'Sending SIGINT to process {self.process.pid}.', sub='error')
		self.process.send_signal(signal.SIGINT)

	def stats(self):
		"""Gather stats about the current running process, if any."""
		if not self.process or not self.process.pid:
			return
		proc = psutil.Process(self.process.pid)
		stats = Command.get_process_info(proc, children=True)
		for info in stats:
			name = info['name']
			pid = info['pid']
			cpu_percent = info['cpu_percent']
			mem_percent = info['memory_percent']
			net_conns = info.get('net_connections') or []
			extra_data = {k: v for k, v in info.items() if k not in ['cpu_percent', 'memory_percent', 'net_connections']}
			yield Stat(
				name=name,
				pid=pid,
				cpu=cpu_percent,
				memory=mem_percent,
				net_conns=len(net_conns),
				extra_data=extra_data
			)

	@staticmethod
	def get_process_info(process, children=False):
		"""Get process information from psutil.

		Args:
			process (subprocess.Process): Process.
			children (bool): Whether to gather stats about children processes too.
		"""
		try:
			data = {
				k: v._asdict() if hasattr(v, '_asdict') else v
				for k, v in process.as_dict().items()
				if k not in ['memory_maps', 'open_files', 'environ']
			}
			yield data
		except (psutil.Error, FileNotFoundError):
			return
		if children:
			for subproc in process.children(recursive=True):
				yield from Command.get_process_info(subproc, children=False)

	def run_item_loaders(self, line):
		"""Run item loaders against an output line.

		Args:
			line (str): Output line.
		"""
		if self.no_process:
			return
		for item_loader in self.item_loaders:
			if (callable(item_loader)):
				yield from item_loader(self, line)
			elif item_loader:
				name = item_loader.__class__.__name__.replace('Serializer', '').lower()
				default_callback = lambda self, x: [(yield x)]  # noqa: E731
				callback = getattr(self, f'on_{name}_loaded', None) or default_callback
				for item in item_loader.run(line):
					yield from callback(self, item)

	def _prompt_sudo(self, command):
		"""
		Checks if the command requires sudo and prompts for the password if necessary.

		Args:
			command (str): The initial command to be executed.

		Returns:
			tuple: (sudo password, error).
		"""
		sudo_password = None

		# Check if sudo is required by the command
		if not re.search(r'\bsudo\b', command):
			return None, []

		# Check if sudo can be executed without a password
		try:
			if subprocess.run(['sudo', '-n', 'true'], capture_output=False).returncode == 0:
				return None, None
		except ValueError:
			self._print('[bold orange3]Could not run sudo check test.[/][bold green]Passing.[/]')

		# Check if we have a tty
		if not os.isatty(sys.stdin.fileno()):
			error = "No TTY detected. Sudo password prompt requires a TTY to proceed."
			return -1, error

		# If not, prompt the user for a password
		self._print('[bold red]Please enter sudo password to continue.[/]', rich=True)
		for _ in range(3):
			user = getpass.getuser()
			self._print(rf'\[sudo] password for {user}: ▌', rich=True)
			sudo_password = getpass.getpass()
			result = subprocess.run(
				['sudo', '-S', '-p', '', 'true'],
				input=sudo_password + "\n",
				text=True,
				capture_output=True
			)
			if result.returncode == 0:
				return sudo_password, None  # Password is correct
			self._print("Sorry, try again.")
		error = "Sudo password verification failed after 3 attempts."
		return -1, error

	def _wait_for_end(self):
		"""Wait for process to finish and process output and return code."""
		if not self.process:
			return
		for line in self.process.stdout.readlines():
			yield from self.process_line(line)
		self.process.wait()
		self.return_code = self.process.returncode
		self.process.stdout.close()
		self.return_code = 0 if self.ignore_return_code else self.return_code
		self.output = self.output.strip()
		self.killed = self.return_code == -2 or self.killed
		self.debug(f'Command {self.cmd} finished with return code {self.return_code}', sub='command')

		if self.killed:
			error = 'Process was killed manually (CTRL+C / CTRL+X)'
			yield Error(
				message=error,
				_source=self.unique_name,
				_uuid=str(uuid.uuid4())
			)

		elif self.return_code != 0:
			error = f'Command failed with return code {self.return_code}'
			last_lines = self.output.split('\n')
			last_lines = last_lines[max(0, len(last_lines) - 2):]
			last_lines = [line for line in last_lines if line != '']
			yield Error(
				message=error,
				traceback='\n'.join(last_lines),
				traceback_title='Last stdout lines',
				_source=self.unique_name,
				_uuid=str(uuid.uuid4())
			)

	@staticmethod
	def _process_opts(
			opts,
			opts_conf,
			opt_key_map={},
			opt_value_map={},
			opt_prefix='-',
			command_name=None):
		"""Process a dict of options using a config, option key map / value map and option character like '-' or '--'.

		Args:
			opts (dict): Command options as input on the CLI.
			opts_conf (dict): Options config (Click options definition).
			opt_key_map (dict[str, str | Callable]): A dict to map option key with their actual values.
			opt_value_map (dict, str | Callable): A dict to map option values with their actual values.
			opt_prefix (str, default: '-'): Option prefix.
			command_name (str | None, default: None): Command name.
		"""
		opts_str = ''
		for opt_name, opt_conf in opts_conf.items():
			debug('before get_opt_value', obj={'name': opt_name, 'conf': opt_conf}, obj_after=False, sub='command.options', verbose=True)  # noqa: E501

			# Get opt value
			default_val = opt_conf.get('default')
			opt_val = Command._get_opt_value(
				opts,
				opt_name,
				opts_conf,
				opt_prefix=command_name,
				default=default_val)

			debug('after get_opt_value', obj={'name': opt_name, 'value': opt_val, 'conf': opt_conf}, obj_after=False, sub='command.options', verbose=True)  # noqa: E501

			# Skip option if value is falsy
			if opt_val in [None, False, []]:
				debug('skipped (falsy)', obj={'name': opt_name, 'value': opt_val}, obj_after=False, sub='command.options', verbose=True)  # noqa: E501
				continue

			# Apply process function on opt value
			if 'process' in opt_conf:
				func = opt_conf['process']
				opt_val = func(opt_val)

			# Convert opt value to expected command opt value
			mapped_opt_val = opt_value_map.get(opt_name)
			if mapped_opt_val:
				if callable(mapped_opt_val):
					opt_val = mapped_opt_val(opt_val)
				else:
					opt_val = mapped_opt_val

			# Convert opt name to expected command opt name
			mapped_opt_name = opt_key_map.get(opt_name)
			if mapped_opt_name is not None:
				if mapped_opt_name == OPT_NOT_SUPPORTED:
					debug('skipped (unsupported)', obj={'name': opt_name, 'value': opt_val}, sub='command.options', verbose=True)  # noqa: E501
					continue
				else:
					opt_name = mapped_opt_name
			debug('mapped key / value', obj={'name': opt_name, 'value': opt_val}, obj_after=False, sub='command.options', verbose=True)  # noqa: E501

			# Avoid shell injections and detect opt prefix
			opt_name = str(opt_name).split(' ')[0]  # avoid cmd injection

			# Replace '_' with '-'
			opt_name = opt_name.replace('_', '-')

			# Add opt prefix if not already there
			if len(opt_name) > 0 and opt_name[0] not in ['-', '--']:
				opt_name = f'{opt_prefix}{opt_name}'

			# Append opt name + opt value to option string.
			# Note: does not append opt value if value is True (flag)
			opts_str += f' {opt_name}'
			shlex_quote = opt_conf.get('shlex', True)
			if opt_val is not True:
				if shlex_quote:
					opt_val = shlex.quote(str(opt_val))
				opts_str += f' {opt_val}'
			debug('final', obj={'name': opt_name, 'value': opt_val}, sub='command.options', obj_after=False, verbose=True)

		return opts_str.strip()

	@staticmethod
	def _validate_chunked_input(self, inputs):
		"""Command does not suport multiple inputs in non-worker mode. Consider using .delay() instead."""
		if len(inputs) > 1 and self.sync and self.file_flag is None:
			return False
		return True

	@staticmethod
	def _validate_input_nonempty(self, inputs):
		"""Input is empty."""
		if not self.input_required:
			return True
		if not inputs or len(inputs) == 0:
			return False
		return True

	# @staticmethod
	# def _validate_input_types_valid(self, input):
	# 	pass

	@staticmethod
	def _get_opt_default(opt_name, opts_conf):
		for k, v in opts_conf.items():
			if k == opt_name:
				return v.get('default', None)
		return None

	@staticmethod
	def _get_opt_value(opts, opt_name, opts_conf={}, opt_prefix='', default=None):
		default = default or Command._get_opt_default(opt_name, opts_conf)
		opt_names = [
			f'{opt_prefix}.{opt_name}',
			f'{opt_prefix}_{opt_name}',
			opt_name,
		]
		opt_values = [opts.get(o) for o in opt_names]
		alias = [conf.get('short') for _, conf in opts_conf.items() if conf.get('short') in opts and _ == opt_name]
		if alias:
			opt_values.append(opts.get(alias[0]))
		if OPT_NOT_SUPPORTED in opt_values:
			debug('skipped (unsupported)', obj={'name': opt_name}, obj_after=False, sub='command.options', verbose=True)
			return None
		value = next((v for v in opt_values if v is not None), default)
		debug('got opt value', obj={'name': opt_name, 'value': value, 'aliases': opt_names, 'values': opt_values}, obj_after=False, sub='command.options', verbose=True)  # noqa: E501
		return value

	def _build_cmd(self):
		"""Build command string."""

		# Add JSON flag to cmd
		if self.json_flag:
			self.cmd += f' {self.json_flag}'

		# Add options to cmd
		opts_str = Command._process_opts(
			self.run_opts,
			self.opts,
			self.opt_key_map,
			self.opt_value_map,
			self.opt_prefix,
			command_name=self.config.name)
		if opts_str:
			self.cmd += f' {opts_str}'

		# Add meta options to cmd
		meta_opts_str = Command._process_opts(
			self.run_opts,
			self.meta_opts,
			self.opt_key_map,
			self.opt_value_map,
			self.opt_prefix,
			command_name=self.config.name)
		if meta_opts_str:
			self.cmd += f' {meta_opts_str}'

	def _build_cmd_input(self):
		"""Many commands take as input a string or a list. This function facilitate this based on whether we pass a
		string or a list to the cmd.
		"""
		cmd = self.cmd
		inputs = self.inputs

		# If inputs is empty, return the previous command
		if not inputs:
			return

		# If inputs has a single element but the tool does not support an input flag, use echo-piped_input input.
		# If the tool's input flag is set to None, assume it is a positional argument at the end of the command.
		# Otherwise use the input flag to pass the input.
		if len(inputs) == 1:
			input = shlex.quote(inputs[0])
			if self.input_flag == OPT_PIPE_INPUT:
				cmd = f'echo {input} | {cmd}'
			elif not self.input_flag:
				cmd += f' {input}'
			else:
				cmd += f' {self.input_flag} {input}'

		# If inputs has multiple elements and the tool has input_flag set to OPT_PIPE_INPUT, use cat-piped_input input.
		# Otherwise pass the file path to the tool.
		else:
			fpath = f'{self.reports_folder}/.inputs/{self.unique_name}.txt'

			# Write the input to a file
			with open(fpath, 'w') as f:
				f.write('\n'.join(inputs))

			if self.file_flag == OPT_PIPE_INPUT:
				cmd = f'cat {fpath} | {cmd}'
			elif self.file_flag:
				cmd += f' {self.file_flag} {fpath}'
			else:
				cmd += f' {fpath}'

			self.inputs_path = fpath

		self.cmd = cmd
		self.shell = ' | ' in self.cmd



import distro
import getpass
import os
import platform
import re
import shutil
import tarfile
import zipfile
import io

from dataclasses import dataclass
from datetime import datetime
from enum import Enum

import json
import requests

from rich.table import Table

from secator.config import CONFIG
from secator.celery import IN_CELERY_WORKER_PROCESS
from secator.definitions import OPT_NOT_SUPPORTED
from secator.output_types import Info, Warning, Error
from secator.rich import console
from secator.runners import Command


class InstallerStatus(Enum):
	SUCCESS = 'SUCCESS'
	INSTALL_FAILED = 'INSTALL_FAILED'
	INSTALL_NOT_SUPPORTED = 'INSTALL_NOT_SUPPORTED'
	INSTALL_SKIPPED_OK = 'INSTALL_SKIPPED_OK'
	GITHUB_LATEST_RELEASE_NOT_FOUND = 'GITHUB_LATEST_RELEASE_NOT_FOUND'
	GITHUB_RELEASE_NOT_FOUND = 'RELEASE_NOT_FOUND'
	GITHUB_RELEASE_FAILED_DOWNLOAD = 'GITHUB_RELEASE_FAILED_DOWNLOAD'
	GITHUB_BINARY_NOT_FOUND_IN_ARCHIVE = 'GITHUB_BINARY_NOT_FOUND_IN_ARCHIVE'
	UNKNOWN_DISTRIBUTION = 'UNKNOWN_DISTRIBUTION'
	UNKNOWN = 'UNKNOWN'

	def is_ok(self):
		return self.value in ['SUCCESS', 'INSTALL_SKIPPED_OK']


@dataclass
class Distribution:
	name: str
	pm_name: str
	pm_installer: str
	pm_finalizer: str


class ToolInstaller:
	status = InstallerStatus

	@classmethod
	def install(cls, tool_cls):
		name = tool_cls.__name__
		console.print(Info(message=f'Installing {name}'))
		status = InstallerStatus.UNKNOWN

		# Fail if not supported
		if not any(_ for _ in [
			tool_cls.install_pre,
			tool_cls.install_github_handle,
			tool_cls.install_cmd,
			tool_cls.install_post]):
			return InstallerStatus.INSTALL_NOT_SUPPORTED

		# Check PATH
		path_var = os.environ.get('PATH', '')
		if not str(CONFIG.dirs.bin) in path_var:
			console.print(Warning(message=f'Bin directory {CONFIG.dirs.bin} not found in PATH ! Binaries installed by secator will not work'))  # noqa: E501
			console.print(Warning(message=f'Run "export PATH=$PATH:{CONFIG.dirs.bin}" to add the binaries to your PATH'))

		# Install pre-required packages
		if tool_cls.install_pre:
			status = PackageInstaller.install(tool_cls.install_pre)
			if not status.is_ok():
				cls.print_status(status, name)
				return status

		# Install binaries from GH
		gh_status = InstallerStatus.UNKNOWN
		if tool_cls.install_github_handle and not CONFIG.security.force_source_install:
			gh_status = GithubInstaller.install(tool_cls.install_github_handle)
			status = gh_status

		# Install from source
		if tool_cls.install_cmd and not gh_status.is_ok():
			status = SourceInstaller.install(tool_cls.install_cmd)
			if not status.is_ok():
				cls.print_status(status, name)
				return status

		# Install post commands
		if tool_cls.install_post:
			post_status = SourceInstaller.install(tool_cls.install_post)
			if not post_status.is_ok():
				cls.print_status(post_status, name)
				return post_status

		cls.print_status(status, name)
		return status

	@classmethod
	def print_status(cls, status, name):
		if status.is_ok():
			console.print(Info(message=f'{name} installed successfully!'))
		elif status == InstallerStatus.INSTALL_NOT_SUPPORTED:
			console.print(Error(message=f'{name} install is not supported yet. Please install manually'))
		else:
			console.print(Error(message=f'Failed to install {name}: {status}'))


class PackageInstaller:
	"""Install system packages."""

	@classmethod
	def install(cls, config):
		"""Install packages using the correct package manager based on the distribution.

		Args:
			config (dict): A dict of package managers as keys and a list of package names as values.

		Returns:
			InstallerStatus: installer status.
		"""
		# Init status
		distribution = get_distro_config()
		if not distribution.pm_installer:
			return InstallerStatus.UNKNOWN_DISTRIBUTION

		console.print(
			Info(message=f'Detected distribution "{distribution.name}", using package manager "{distribution.pm_name}"'))

		# Construct package list
		pkg_list = []
		for managers, packages in config.items():
			if distribution.pm_name in managers.split("|") or managers == '*':
				pkg_list.extend(packages)
				break

		# Installer cmd
		cmd = distribution.pm_installer
		if CONFIG.security.autoinstall_commands and IN_CELERY_WORKER_PROCESS:
			cmd = f'flock /tmp/install.lock {cmd}'
		if getpass.getuser() != 'root':
			cmd = f'sudo {cmd}'

		if pkg_list:
			pkg_str = ''
			for pkg in pkg_list:
				if ':' in pkg:
					pdistro, pkg = pkg.split(':')
					if pdistro != distribution.name:
						continue
				pkg_str += f'{pkg} '
			console.print(Info(message=f'Installing packages {pkg_str}'))
			status = SourceInstaller.install(f'{cmd} {pkg_str}', install_prereqs=False)
			if not status.is_ok():
				return status
		return InstallerStatus.SUCCESS


class SourceInstaller:
	"""Install a tool from source."""

	@classmethod
	def install(cls, config, install_prereqs=True):
		"""Install from source.

		Args:
			cls: ToolInstaller class.
			config (dict): A dict of distros as keys and a command as value.

		Returns:
			Status: install status.
		"""
		install_cmd = None
		if isinstance(config, str):
			install_cmd = config
		else:
			distribution = get_distro_config()
			for distros, command in config.items():
				if distribution.name in distros.split("|") or distros == '*':
					install_cmd = command
					break
		if not install_cmd:
			return InstallerStatus.INSTALL_SKIPPED_OK

		# Install build dependencies if needed
		if install_prereqs:
			if 'go ' in install_cmd:
				status = PackageInstaller.install({'apt': ['golang-go'], '*': ['go']})
				if not status.is_ok():
					return status
			if 'gem ' in install_cmd:
				status = PackageInstaller.install({'apk': ['ruby', 'ruby-dev'], 'pacman': ['ruby', 'rubygems'], 'apt': ['ruby-full', 'rubygems']})  # noqa: E501
				if not status.is_ok():
					return status
			if 'git ' in install_cmd or 'git+' in install_cmd:
				status = PackageInstaller.install({'*': ['git']})
				if not status.is_ok():
					return status

		# Run command
		ret = Command.execute(install_cmd, cls_attributes={'shell': True}, quiet=False)
		return InstallerStatus.SUCCESS if ret.return_code == 0 else InstallerStatus.INSTALL_FAILED


class GithubInstaller:
	"""Install a tool from GitHub releases."""

	@classmethod
	def install(cls, github_handle):
		"""Find and install a release from a GitHub handle {user}/{repo}.

		Args:
			github_handle (str): A GitHub handle {user}/{repo}

		Returns:
			InstallerStatus: status.
		"""
		_, repo = tuple(github_handle.split('/'))
		latest_release = cls.get_latest_release(github_handle)
		if not latest_release:
			return InstallerStatus.GITHUB_LATEST_RELEASE_NOT_FOUND

		# Find the right asset to download
		os_identifiers, arch_identifiers = cls._get_platform_identifier()
		download_url = cls._find_matching_asset(latest_release['assets'], os_identifiers, arch_identifiers)
		if not download_url:
			console.print(Error(message='Could not find a GitHub release matching distribution.'))
			return InstallerStatus.GITHUB_RELEASE_NOT_FOUND

		# Download and unpack asset
		console.print(Info(message=f'Found release URL: {download_url}'))
		return cls._download_and_unpack(download_url, CONFIG.dirs.bin, repo)

	@classmethod
	def get_latest_release(cls, github_handle):
		"""Get latest release from GitHub.

		Args:
			github_handle (str): A GitHub handle {user}/{repo}.

		Returns:
			dict: Latest release JSON from GitHub releases.
		"""
		if not github_handle:
			return False
		owner, repo = tuple(github_handle.split('/'))
		url = f"https://api.github.com/repos/{owner}/{repo}/releases/latest"
		headers = {}
		if CONFIG.cli.github_token:
			headers['Authorization'] = f'Bearer {CONFIG.cli.github_token}'
		try:
			response = requests.get(url, headers=headers, timeout=5)
			response.raise_for_status()
			latest_release = response.json()
			return latest_release
		except requests.RequestException as e:
			console.print(Warning(message=f'Failed to fetch latest release for {github_handle}: {str(e)}'))
			return None

	@classmethod
	def get_latest_version(cls, github_handle):
		latest_release = cls.get_latest_release(github_handle)
		if not latest_release:
			return None
		return latest_release['tag_name'].lstrip('v')

	@classmethod
	def _get_platform_identifier(cls):
		"""Generate lists of possible identifiers for the current platform."""
		system = platform.system().lower()
		arch = platform.machine().lower()

		# Mapping common platform.system() values to those found in release names
		os_mapping = {
			'linux': ['linux'],
			'windows': ['windows', 'win'],
			'darwin': ['darwin', 'macos', 'osx', 'mac']
		}

		# Enhanced architecture mapping to avoid conflicts
		arch_mapping = {
			'x86_64': ['amd64', 'x86_64'],
			'amd64': ['amd64', 'x86_64'],
			'aarch64': ['arm64', 'aarch64'],
			'armv7l': ['armv7', 'arm'],
			'386': ['386', 'x86', 'i386'],
		}

		os_identifiers = os_mapping.get(system, [])
		arch_identifiers = arch_mapping.get(arch, [])
		return os_identifiers, arch_identifiers

	@classmethod
	def _find_matching_asset(cls, assets, os_identifiers, arch_identifiers):
		"""Find a release asset matching the current platform more precisely."""
		potential_matches = []

		for asset in assets:
			asset_name = asset['name'].lower()
			if any(os_id in asset_name for os_id in os_identifiers) and \
			   any(arch_id in asset_name for arch_id in arch_identifiers):
				potential_matches.append(asset['browser_download_url'])

		# Preference ordering for file formats, if needed
		preferred_formats = ['.tar.gz', '.zip']

		for format in preferred_formats:
			for match in potential_matches:
				if match.endswith(format):
					return match

		if potential_matches:
			return potential_matches[0]

	@classmethod
	def _download_and_unpack(cls, url, destination, repo_name):
		"""Download and unpack a release asset.

		Args:
			cls (Runner): Task class.
			url (str): GitHub release URL.
			destination (str): Local destination.
			repo_name (str): GitHub repository name.

		Returns:
			InstallerStatus: install status.
		"""
		console.print(Info(message=f'Downloading and unpacking to {destination}...'))
		response = requests.get(url, timeout=5)
		if not response.status_code == 200:
			return InstallerStatus.GITHUB_RELEASE_FAILED_DOWNLOAD

		# Create a temporary directory to extract the archive
		date_str = datetime.now().strftime("%Y%m%d_%H%M%S")
		temp_dir = os.path.join("/tmp", f'{repo_name}_{date_str}')
		os.makedirs(temp_dir, exist_ok=True)

		console.print(Info(message=f'Extracting binary to {temp_dir}...'))
		if url.endswith('.zip'):
			with zipfile.ZipFile(io.BytesIO(response.content)) as zip_ref:
				zip_ref.extractall(temp_dir)
		elif url.endswith('.tar.gz'):
			with tarfile.open(fileobj=io.BytesIO(response.content), mode='r:gz') as tar:
				tar.extractall(path=temp_dir)

		# For archives, find and move the binary that matches the repo name
		binary_path = cls._find_binary_in_directory(temp_dir, repo_name)
		if binary_path:
			os.chmod(binary_path, 0o755)  # Make it executable
			destination = os.path.join(destination, repo_name)
			console.print(Info(message=f'Moving binary to {destination}...'))
			shutil.move(binary_path, destination)  # Move the binary
			return InstallerStatus.SUCCESS
		else:
			console.print(Error(message='Binary matching the repository name was not found in the archive.'))
			return InstallerStatus.GITHUB_BINARY_NOT_FOUND_IN_ARCHIVE

	@classmethod
	def _find_binary_in_directory(cls, directory, binary_name):
		"""Search for the binary in the given directory that matches the repository name."""
		for root, _, files in os.walk(directory):
			for file in files:
				# Match the file name exactly with the repository name
				if file == binary_name:
					return os.path.join(root, file)
		return None


def which(command):
	"""Run which on a command.

	Args:
		command (str): Command to check.

	Returns:
		secator.Command: Command instance.
	"""
	return Command.execute(f'which {command}', quiet=True, print_errors=False)


def get_version(version_cmd):
	"""Run version command and match first version number found.

	Args:
		version_cmd (str): Command to get the version.

	Returns:
		tuple[str]: Version string, return code.
	"""
	from secator.runners import Command
	import re
	regex = r'[0-9]+\.[0-9]+\.?[0-9]*\.?[a-zA-Z]*'
	ret = Command.execute(version_cmd, quiet=True, print_errors=False)
	return_code = ret.return_code
	if not return_code == 0:
		return '', ret.return_code
	match = re.findall(regex, ret.output)
	if not match:
		return '', return_code
	return match[0], return_code


def parse_version(ver):
	from packaging import version as _version
	try:
		return _version.parse(ver)
	except _version.InvalidVersion:
		version_regex = re.compile(r'(\d+\.\d+(?:\.\d+)?)')
		match = version_regex.search(ver)
		if match:
			return _version.parse(match.group(1))
		return None


def get_version_info(name, version_flag=None, install_github_handle=None, install_cmd=None, version=None):
	"""Get version info for a command.

	Args:
		name (str): Command name.
		version_flag (str): Version flag.
		install_github_handle (str): Github handle.
		install_cmd (str): Install command.
		version (str): Existing version.

	Return:
		dict: Version info.
	"""
	from secator.installer import GithubInstaller
	info = {
		'name': name,
		'installed': False,
		'version': version,
		'latest_version': None,
		'location': None,
		'status': ''
	}

	# Get binary path
	location = which(name).output
	info['location'] = location

	# Get latest version
	latest_version = None
	if not CONFIG.offline_mode:
		if install_github_handle:
			latest_version = GithubInstaller.get_latest_version(install_github_handle)
			info['latest_version'] = latest_version
		elif install_cmd and install_cmd.startswith('pip'):
			req = requests.get(f'https://pypi.python.org/pypi/{name}/json')
			version = parse_version('0')
			if req.status_code == requests.codes.ok:
				j = json.loads(req.text.encode(req.encoding))
				releases = j.get('releases', [])
				for release in releases:
					ver = parse_version(release)
					if ver and not ver.is_prerelease:
						version = max(version, ver)
						latest_version = str(version)
						info['latest_version'] = latest_version
		elif install_cmd and install_cmd.startswith('sudo apt install'):
			ret = Command.execute(f'apt-cache madison {name}', quiet=True)
			if ret.return_code == 0:
				output = ret.output.split(' | ')
				if len(output) > 1:
					ver = parse_version(output[1].strip())
					if ver:
						latest_version = str(ver)
						info['latest_version'] = latest_version

	# Get current version
	version_ret = 1
	version_flag = None if version_flag == OPT_NOT_SUPPORTED else version_flag
	if version_flag:
		version_cmd = f'{name} {version_flag}'
		version, version_ret = get_version(version_cmd)
		info['version'] = version
		if version_ret != 0:  # version command error
			info['installed'] = False
			info['status'] = 'missing'
			return info

	if location:
		info['installed'] = True
		if version and latest_version:
			if parse_version(version) < parse_version(latest_version):
				info['status'] = 'outdated'
			else:
				info['status'] = 'latest'
		elif not version:
			info['status'] = 'current unknown'
		elif not latest_version:
			info['status'] = 'latest unknown'
			if CONFIG.offline_mode:
				info['status'] += r' [dim orange1]\[offline][/]'
	else:
		info['status'] = 'missing'

	return info


def get_distro_config():
	"""Detects the system's package manager based on the OS distribution and return the default installation command."""

	# If explicitely set by the user, use that one
	package_manager_variable = os.environ.get('SECATOR_PACKAGE_MANAGER')
	if package_manager_variable:
		return package_manager_variable
	installer = None
	finalizer = None
	system = platform.system()
	distrib = system

	if system == "Linux":
		distrib = distro.id()

		if distrib in ["ubuntu", "debian", "linuxmint", "popos", "kali"]:
			installer = "apt install -y --no-install-recommends"
			finalizer = "rm -rf /var/lib/apt/lists/*"
		elif distrib in ["arch", "manjaro", "endeavouros"]:
			installer = "pacman -S --noconfirm --needed"
		elif distrib in ["alpine"]:
			installer = "apk add --no-cache"
		elif distrib in ["fedora"]:
			installer = "dnf install -y"
			finalizer = "dnf clean all"
		elif distrib in ["centos", "rhel", "rocky", "alma"]:
			installer = "yum -y"
			finalizer = "yum clean all"
		elif distrib in ["opensuse", "sles"]:
			installer = "zypper -n"
			finalizer = "zypper clean --all"

	elif system == "Darwin":  # macOS
		installer = "brew install"

	elif system == "Windows":
		if shutil.which("winget"):
			installer = "winget install --disable-interactivity"
		elif shutil.which("choco"):
			installer = "choco install -y --no-progress"
		else:
			installer = "scoop"  # Alternative package manager for Windows

	manager = installer.split(' ')[0]
	config = Distribution(
		pm_installer=installer,
		pm_finalizer=finalizer,
		pm_name=manager,
		name=distrib
	)
	return config


def fmt_health_table_row(version_info, category=None):
	name = version_info['name']
	version = version_info['version']
	status = version_info['status']
	installed = version_info['installed']
	latest_version = version_info['latest_version']
	name_str = f'[magenta]{name:<13}[/]'

	# Format version row
	_version = version or ''
	_version = f'[bold green]{_version:<10}[/]'
	if status == 'latest':
		_version += ' [bold green](latest)[/]'
	elif status == 'outdated':
		_version += ' [bold red](outdated)[/]'
		if latest_version:
			_version += f' [dim](<{latest_version})'
	elif status == 'missing':
		_version = '[bold red]missing[/]'
	elif status == 'ok':
		_version = '[bold green]ok        [/]'
	elif status:
		if not version and installed:
			_version = '[bold green]ok        [/]'
		_version += f' [dim]({status}[/])'

	row = (name_str, _version)
	return row


def get_health_table():
	table = Table(box=None, show_header=False)
	for col in ['name', 'version']:
		table.add_column(col)
	return table



import json
import os
import re

from functools import cache

import requests
from bs4 import BeautifulSoup
from cpe import CPE

from secator.definitions import (CIDR_RANGE, CVSS_SCORE, DELAY, DEPTH, DESCRIPTION, FILTER_CODES,
								 FILTER_REGEX, FILTER_SIZE, FILTER_WORDS, FOLLOW_REDIRECT, HEADER, HOST, ID, IP,
								 MATCH_CODES, MATCH_REGEX, MATCH_SIZE, MATCH_WORDS, METHOD, NAME, PATH, PROVIDER, PROXY,
								 RATE_LIMIT, REFERENCES, RETRIES, SEVERITY, TAGS, THREADS, TIMEOUT, URL, USER_AGENT,
								 USERNAME, WORDLIST)
from secator.output_types import Ip, Port, Subdomain, Tag, Url, UserAccount, Vulnerability
from secator.config import CONFIG
from secator.runners import Command
from secator.utils import debug, process_wordlist


OPTS = {
	HEADER: {'type': str, 'help': 'Custom header to add to each request in the form "KEY1:VALUE1; KEY2:VALUE2"'},
	DELAY: {'type': float, 'short': 'd', 'help': 'Delay to add between each requests'},
	DEPTH: {'type': int, 'help': 'Scan depth', 'default': 2},
	FILTER_CODES: {'type': str, 'short': 'fc', 'help': 'Filter out responses with HTTP codes'},
	FILTER_REGEX: {'type': str, 'short': 'fr', 'help': 'Filter out responses with regular expression'},
	FILTER_SIZE: {'type': str, 'short': 'fs', 'help': 'Filter out responses with size'},
	FILTER_WORDS: {'type': str, 'short': 'fw', 'help': 'Filter out responses with word count'},
	FOLLOW_REDIRECT: {'is_flag': True, 'short': 'frd', 'help': 'Follow HTTP redirects'},
	MATCH_CODES: {'type': str, 'short': 'mc', 'help': 'Match HTTP status codes e.g "201,300,301"'},
	MATCH_REGEX: {'type': str, 'short': 'mr', 'help': 'Match responses with regular expression'},
	MATCH_SIZE: {'type': str, 'short': 'ms', 'help': 'Match respones with size'},
	MATCH_WORDS: {'type': str, 'short': 'mw', 'help': 'Match responses with word count'},
	METHOD: {'type': str, 'help': 'HTTP method to use for requests'},
	PROXY: {'type': str, 'help': 'HTTP(s) / SOCKS5 proxy'},
	RATE_LIMIT: {'type':  int, 'short': 'rl', 'help': 'Rate limit, i.e max number of requests per second'},
	RETRIES: {'type': int, 'help': 'Retries'},
	THREADS: {'type': int, 'help': 'Number of threads to run', 'default': 50},
	TIMEOUT: {'type': int, 'help': 'Request timeout'},
	USER_AGENT: {'type': str, 'short': 'ua', 'help': 'User agent, e.g "Mozilla Firefox 1.0"'},
	WORDLIST: {'type': str, 'short': 'w', 'default': 'http', 'process': process_wordlist, 'help': 'Wordlist to use'}
}

OPTS_HTTP = [
	HEADER, DELAY, FOLLOW_REDIRECT, METHOD, PROXY, RATE_LIMIT, RETRIES, THREADS, TIMEOUT, USER_AGENT
]

OPTS_HTTP_CRAWLERS = OPTS_HTTP + [
	DEPTH, MATCH_REGEX, MATCH_SIZE, MATCH_WORDS, FILTER_REGEX, FILTER_CODES, FILTER_SIZE, FILTER_WORDS,
	MATCH_CODES
]

OPTS_HTTP_FUZZERS = OPTS_HTTP_CRAWLERS + [WORDLIST]

OPTS_RECON = [
	DELAY, PROXY, RATE_LIMIT, RETRIES, THREADS, TIMEOUT
]

OPTS_VULN = [
	HEADER, DELAY, FOLLOW_REDIRECT, PROXY, RATE_LIMIT, RETRIES, THREADS, TIMEOUT, USER_AGENT
]


#---------------#
# HTTP category #
#---------------#

class Http(Command):
	meta_opts = {k: OPTS[k] for k in OPTS_HTTP_CRAWLERS}
	input_type = URL
	output_types = [Url]


class HttpCrawler(Command):
	meta_opts = {k: OPTS[k] for k in OPTS_HTTP_CRAWLERS}
	input_type = URL
	output_types = [Url]


class HttpFuzzer(Command):
	meta_opts = {k: OPTS[k] for k in OPTS_HTTP_FUZZERS}
	input_type = URL
	output_types = [Url]


#----------------#
# Recon category #
#----------------#

class Recon(Command):
	meta_opts = {k: OPTS[k] for k in OPTS_RECON}
	output_types = [Subdomain, UserAccount, Ip, Port]


class ReconDns(Recon):
	input_type = HOST
	output_types = [Subdomain]


class ReconUser(Recon):
	input_type = USERNAME
	output_types = [UserAccount]


class ReconIp(Recon):
	input_type = CIDR_RANGE
	output_types = [Ip]


class ReconPort(Recon):
	input_type = IP
	output_types = [Port]


#---------------#
# Vuln category #
#---------------#

class Vuln(Command):
	meta_opts = {k: OPTS[k] for k in OPTS_VULN}
	output_types = [Vulnerability]

	@staticmethod
	def lookup_local_cve(cve_id):
		cve_path = f'{CONFIG.dirs.data}/cves/{cve_id}.json'
		if os.path.exists(cve_path):
			with open(cve_path, 'r') as f:
				return json.load(f)
		debug(f'CVE {cve_id} not found in cache', sub='cve')
		return None

	# @staticmethod
	# def lookup_exploitdb(exploit_id):
	# 	print('looking up exploit')
	# 	try:
	# 		resp = requests.get(f'https://exploit-db.com/exploits/{exploit_id}', timeout=5)
	#		resp.raise_for_status()
	#		content = resp.content
	# 	except requests.RequestException as e:
	#		debug(f'Failed remote query for {exploit_id} ({str(e)}).', sub='cve')
	# 		logger.error(f'Could not fetch exploit info for exploit {exploit_id}. Skipping.')
	# 		return None
	# 	return cve_info

	@staticmethod
	def create_cpe_string(product_name, version):
		"""
		Generate a CPE string for a given product and version.

		Args:
			product_name (str): The name of the product.
			version (str): The version of the product.

		Returns:
			str: A CPE string formatted according to the CPE 2.3 specification.
		"""
		cpe_version = "2.3"  # CPE Specification version
		part = "a"           # 'a' for application
		vendor = product_name.lower()  # Vendor name, using product name
		product = product_name.lower()  # Product name
		version = version  # Product version
		cpe_string = f"cpe:{cpe_version}:{part}:{vendor}:{product}:{version}:*:*:*:*:*:*:*"
		return cpe_string

	@staticmethod
	def match_cpes(fs1, fs2):
		"""Check if two CPEs match. Partial matches consisting of <vendor>:<product>:<version> are considered a match.

		Args:
			fs1 (str): Format string 1.
			fs2 (str): Format string 2.

		Returns:
			bool: True if the two CPEs match, False otherwise.
		"""
		if fs1 == fs2:
			return True
		split_fs1 = fs1.split(':')
		split_fs2 = fs2.split(':')
		tup1 = split_fs1[3], split_fs1[4], split_fs1[5]
		tup2 = split_fs2[3], split_fs2[4], split_fs2[5]
		return tup1 == tup2

	@staticmethod
	def get_cpe_fs(cpe):
		""""Return formatted string for given CPE.

		Args:
			cpe (string): Input CPE

		Returns:
			string: CPE formatted string.
		"""
		try:
			return CPE(cpe).as_fs()
		except NotImplementedError:
			return None

	@cache
	@staticmethod
	def lookup_cve_from_vulners_exploit(exploit_id, *cpes):
		"""Search for a CVE corresponding to an exploit by extracting the CVE id from the exploit HTML page.

		Args:
			exploit_id (str): Exploit ID.
			cpes (tuple[str], Optional): CPEs to match for.

		Returns:
			dict: vulnerability data.
		"""
		if CONFIG.runners.skip_exploit_search:
			debug(f'Skip remote query for {exploit_id} since config.runners.skip_exploit_search is set.', sub='cve')
			return None
		if CONFIG.offline_mode:
			debug(f'Skip remote query for {exploit_id} since config.offline_mode is set.', sub='cve')
			return None
		try:
			resp = requests.get(f'https://vulners.com/githubexploit/{exploit_id}', timeout=5)
			resp.raise_for_status()
			soup = BeautifulSoup(resp.text, 'lxml')
			title = soup.title.get_text(strip=True)
			h1 = [h1.get_text(strip=True) for h1 in soup.find_all('h1')]
			if '404' in h1:
				raise requests.RequestException("404 [not found or rate limited]")
			code = [code.get_text(strip=True) for code in soup.find_all('code')]
			elems = [title] + h1 + code
			content = '\n'.join(elems)
			cve_regex = re.compile(r'(CVE(?:-|_)\d{4}(?:-|_)\d{4,7})', re.IGNORECASE)
			matches = cve_regex.findall(str(content))
			if not matches:
				debug(f'{exploit_id}: No CVE found in https://vulners.com/githubexploit/{exploit_id}.', sub='cve')
				return None
			cve_id = matches[0].replace('_', '-').upper()
			cve_data = Vuln.lookup_cve(cve_id, *cpes)
			if cve_data:
				return cve_data

		except requests.RequestException as e:
			debug(f'Failed remote query for {exploit_id} ({str(e)}).', sub='cve')
			return None

	@cache
	@staticmethod
	def lookup_cve_from_cve_circle(cve_id):
		"""Get CVE data from vulnerability.circl.lu.

		Args:
			cve_id (str): CVE id.

		Returns:
			dict | None: CVE data, None if no response or empty response.
		"""
		try:
			resp = requests.get(f'https://vulnerability.circl.lu/api/cve/{cve_id}', timeout=5)
			resp.raise_for_status()
			cve_info = resp.json()
			if not cve_info:
				debug(f'Empty response from https://vulnerability.circl.lu/api/cve/{cve_id}', sub='cve')
				return None
			cve_path = f'{CONFIG.dirs.data}/cves/{cve_id}.json'
			with open(cve_path, 'w') as f:
				f.write(json.dumps(cve_info, indent=2))
			debug(f'Downloaded {cve_id} to {cve_path}', sub='cve')
			return cve_info
		except requests.RequestException as e:
			debug(f'Failed remote query for {cve_id} ({str(e)}).', sub='cve')
			return None

	@cache
	@staticmethod
	def lookup_cve(cve_id, *cpes):
		"""Search for a CVE info and return vulnerability data.

		Args:
			cve_id (str): CVE ID in the form CVE-*
			cpes (tuple[str], Optional): CPEs to match for.

		Returns:
			dict: vulnerability data.
		"""
		cve_info = Vuln.lookup_local_cve(cve_id)

		# Online CVE lookup
		if not cve_info:
			if CONFIG.runners.skip_cve_search:
				debug(f'Skip remote query for {cve_id} since config.runners.skip_cve_search is set.', sub='cve')
				return None
			if CONFIG.offline_mode:
				debug(f'Skip remote query for {cve_id} since config.offline_mode is set.', sub='cve')
				return None
			cve_info = Vuln.lookup_cve_from_cve_circle(cve_id)
			if not cve_info:
				return None

		# Convert cve info to easy format
		cve_id = cve_info['cveMetadata']['cveId']
		cna = cve_info['containers']['cna']
		metrics = cna.get('metrics', [])
		cvss_score = 0
		for metric in metrics:
			for name, value in metric.items():
				if 'cvss' in name:
					cvss_score = metric[name]['baseScore']
		description = cna.get('descriptions', [{}])[0].get('value')
		cwe_id = cna.get('problemTypes', [{}])[0].get('descriptions', [{}])[0].get('cweId')
		cpes_affected = []
		for product in cna['affected']:
			cpes_affected.extend(product.get('cpes', []))
		references = [u['url'] for u in cna['references']]
		cve_info = {
			'id': cve_id,
			'cwe_id': cwe_id,
			'cvss_score': cvss_score,
			'description': description,
			'cpes': cpes_affected,
			'references': references
		}

		# Match the CPE string against the affected products CPE FS strings from the CVE data if a CPE was passed.
		# This allow to limit the number of False positives (high) that we get from nmap NSE vuln scripts like vulscan
		# and ensure we keep only right matches.
		# The check is not executed if no CPE was passed (sometimes nmap cannot properly detect a CPE) or if the CPE
		# version cannot be determined.
		cpe_match = False
		tags = [cve_id]
		if cpes:
			for cpe in cpes:
				cpe_fs = Vuln.get_cpe_fs(cpe)
				if not cpe_fs:
					debug(f'{cve_id}: Failed to parse CPE {cpe} with CPE parser', sub='cve.match', verbose=True)
					tags.append('cpe-invalid')
					continue
				# cpe_version = cpe_obj.get_version()[0]
				for cpe_affected in cpes_affected:
					cpe_affected_fs = Vuln.get_cpe_fs(cpe_affected)
					if not cpe_affected_fs:
						debug(f'{cve_id}: Failed to parse CPE {cpe} (from online data) with CPE parser', sub='cve.match', verbose=True)
						continue
					debug(f'{cve_id}: Testing {cpe_fs} against {cpe_affected_fs}', sub='cve.match', verbose=True)
					cpe_match = Vuln.match_cpes(cpe_fs, cpe_affected_fs)
					if cpe_match:
						debug(f'{cve_id}: CPE match found for {cpe}.', sub='cve')
						tags.append('cpe-match')
						break

				if not cpe_match:
					debug(f'{cve_id}: no CPE match found for {cpe}.', sub='cve')

		# Parse CVE id and CVSS
		name = id = cve_info['id']
		# exploit_ids = cve_info.get('refmap', {}).get('exploit-db', [])
		# osvdb_ids = cve_info.get('refmap', {}).get('osvdb', [])

		# Get description
		description = cve_info['description']
		if description is not None:
			description = description.replace(id, '').strip()

		# Get references
		references = cve_info.get(REFERENCES, [])
		cve_ref_url = f'https://vulnerability.circl.lu/cve/{id}'
		references.append(cve_ref_url)

		# Get CWE ID
		cwe_id = cve_info['cwe_id']
		if cwe_id is not None:
			tags.append(cwe_id)

		# Set vulnerability severity based on CVSS score
		severity = None
		cvss = cve_info['cvss_score']
		if cvss:
			severity = Vuln.cvss_to_severity(cvss)

		# Set confidence
		vuln = {
			ID: id,
			NAME: name,
			PROVIDER: 'vulnerability.circl.lu',
			SEVERITY: severity,
			CVSS_SCORE: cvss,
			TAGS: tags,
			REFERENCES: [f'https://vulnerability.circl.lu/cve/{id}'] + references,
			DESCRIPTION: description,
		}
		return vuln

	@cache
	@staticmethod
	def lookup_ghsa(ghsa_id):
		"""Search for a GHSA on Github and and return associated CVE vulnerability data.

		Args:
			ghsa (str): CVE ID in the form GHSA-*

		Returns:
			dict: vulnerability data.
		"""
		try:
			resp = requests.get(f'https://github.com/advisories/{ghsa_id}', timeout=5)
			resp.raise_for_status()
		except requests.RequestException as e:
			debug(f'Failed remote query for {ghsa_id} ({str(e)}).', sub='cve')
			return None
		soup = BeautifulSoup(resp.text, 'lxml')
		sidebar_items = soup.find_all('div', {'class': 'discussion-sidebar-item'})
		cve_id = sidebar_items[2].find('div').text.strip()
		vuln = Vuln.lookup_cve(cve_id)
		if vuln:
			vuln[TAGS].append('ghsa')
			return vuln
		return None

	@staticmethod
	def cvss_to_severity(cvss):
		if cvss < 4:
			severity = 'low'
		elif cvss < 7:
			severity = 'medium'
		elif cvss < 9:
			severity = 'high'
		else:
			severity = 'critical'
		return severity


class VulnHttp(Vuln):
	input_type = HOST


class VulnCode(Vuln):
	input_type = PATH


class VulnMulti(Vuln):
	input_type = HOST
	output_types = [Vulnerability]


#--------------#
# Tag category #
#--------------#

class Tagger(Command):
	input_type = URL
	output_types = [Tag]

#----------------#
# osint category #
#----------------#


class OSInt(Command):
	output_types = [UserAccount]



import time
from dataclasses import dataclass, field

from secator.definitions import (CONTENT_LENGTH, CONTENT_TYPE, STATUS_CODE,
								 TECH, TIME, TITLE, URL, WEBSERVER)
from secator.output_types import OutputType
from secator.utils import rich_to_ansi, trim_string, rich_escape as _s
from secator.config import CONFIG


@dataclass
class Url(OutputType):
	url: str
	host: str = field(default='', compare=False)
	status_code: int = field(default=0, compare=False)
	title: str = field(default='', compare=False)
	webserver: str = field(default='', compare=False)
	tech: list = field(default_factory=list, compare=False)
	content_type: str = field(default='', compare=False)
	content_length: int = field(default=0, compare=False)
	time: str = field(default='', compare=False)
	method: str = field(default='', compare=False)
	words: int = field(default=0, compare=False)
	lines: int = field(default=0, compare=False)
	screenshot_path: str = field(default='', compare=False)
	stored_response_path: str = field(default='', compare=False)
	headers: dict = field(default_factory=dict, repr=True, compare=False)
	_source: str = field(default='', repr=True, compare=False)
	_type: str = field(default='url', repr=True)
	_timestamp: int = field(default_factory=lambda: time.time(), compare=False)
	_uuid: str = field(default='', repr=True, compare=False)
	_context: dict = field(default_factory=dict, repr=True, compare=False)
	_tagged: bool = field(default=False, repr=True, compare=False)
	_duplicate: bool = field(default=False, repr=True, compare=False)
	_related: list = field(default_factory=list, compare=False)

	_table_fields = [
		URL,
		STATUS_CODE,
		TITLE,
		WEBSERVER,
		TECH,
		CONTENT_TYPE,
		CONTENT_LENGTH,
		TIME
	]
	_sort_by = (URL,)

	def __gt__(self, other):
		# favor httpx over other url info tools
		if self._source == 'httpx' and other._source != 'httpx':
			return True
		return super().__gt__(other)

	def __str__(self):
		return self.url

	def __repr__(self):
		s = f'🔗 [white]{_s(self.url)}'
		if self.method and self.method != 'GET':
			s += rf' \[[turquoise4]{self.method}[/]]'
		if self.status_code and self.status_code != 0:
			if self.status_code < 400:
				s += rf' \[[green]{self.status_code}[/]]'
			else:
				s += rf' \[[red]{self.status_code}[/]]'
		if self.title:
			s += rf' \[[green]{trim_string(self.title)}[/]]'
		if self.webserver:
			s += rf' \[[magenta]{_s(self.webserver)}[/]]'
		if self.tech:
			techs_str = ', '.join([f'[magenta]{_s(tech)}[/]' for tech in self.tech])
			s += f' [{techs_str}]'
		if self.content_type:
			s += rf' \[[magenta]{_s(self.content_type)}[/]]'
		if self.content_length:
			cl = str(self.content_length)
			cl += '[bold red]+[/]' if self.content_length == CONFIG.http.response_max_size_bytes else ''
			s += rf' \[[magenta]{cl}[/]]'
		if self.screenshot_path:
			s += rf' \[[magenta]{_s(self.screenshot_path)}[/]]'
		return rich_to_ansi(s)



import validators

from secator.decorators import task
from secator.definitions import (DELAY, IP, OPT_NOT_SUPPORTED, PROXY, RATE_LIMIT,
							   RETRIES, THREADS, TIMEOUT)
from secator.output_types import Ip
from secator.tasks._categories import ReconIp


@task()
class fping(ReconIp):
	"""Send ICMP echo probes to network hosts, similar to ping, but much better."""
	cmd = 'fping -a'
	file_flag = '-f'
	input_flag = None
	opt_prefix = '--'
	opt_key_map = {
		DELAY: 'period',
		PROXY: OPT_NOT_SUPPORTED,
		RATE_LIMIT: OPT_NOT_SUPPORTED,
		RETRIES: 'retry',
		TIMEOUT: 'timeout',
		THREADS: OPT_NOT_SUPPORTED
	}
	opt_value_map = {
		DELAY: lambda x: x * 1000,  # convert s to ms
		TIMEOUT: lambda x: x * 1000  # convert s to ms
	}
	input_type = IP
	output_types = [Ip]
	install_pre = {'*': ['fping']}

	@staticmethod
	def item_loader(self, line):
		if not (validators.ipv4(line) or validators.ipv6(line)):
			return
		yield {'ip': line, 'alive': True}

	@staticmethod
	def on_line(self, line):
		if 'Unreachable' in line:
			return ''  # discard line as it pollutes output
		return line



import copy
import json
import logging
import unittest
import unittest.mock
import warnings

from secator.config import CONFIG
from secator.definitions import (DEBUG, DELAY, FOLLOW_REDIRECT, HEADER, HOST,
							   MATCH_CODES, OPT_NOT_SUPPORTED, RATE_LIMIT,
							   THREADS, TIMEOUT)
from secator.output_types import Url
from secator.rich import console
from secator.runners import Command
from secator.tasks import httpx
from secator.utils import setup_logging
from secator.utils_test import (FIXTURES_TASKS, FIXTURES_DIR, INPUTS_TASKS, META_OPTS,
							  TEST_TASKS, CommandOutputTester, load_fixture,
							  mock_command, mock_subprocess_popen)

level = logging.DEBUG if DEBUG > 0 else logging.ERROR
setup_logging(level)


class FakeCmd(Command):
	opts = {
		'opt1': {'type': int, 'default': 10},
		'opt2': {'type': str, 'default': '1,2,3'},
		'opt3': {'is_flag': True, 'default': False}, # optional
		'opt_with_underscore': {'type': str}, # optional
		'opt-with-hyphen': {'type': str}
	}
	opt_prefix = '-' # all options have '-' as option char
	opt_key_map = {
		'opt3': '--opt3', # but opt3 has '--' prefix
		'opt4': OPT_NOT_SUPPORTED
	}
	opt_value_map = {
		'opt1': lambda x: float(x) # actually opt1 value should be cast to a float
	}


class TestCommandProcessOpts(unittest.TestCase):
	
	def setUp(self):
		self.maxDiff = None

	def test_process_opts_defaults(self):
		run_opts = {}
		opts_str = FakeCmd._process_opts(
			run_opts,
			FakeCmd.opts,
			FakeCmd.opt_key_map,
			FakeCmd.opt_value_map)
		self.assertEqual(opts_str, '-opt1 10.0 -opt2 1,2,3')

	def test_process_opts(self):
		run_opts = {
			'opt1': 41,
			'opt2': False, # intentionally omit arg, overriding default value
			'opt3': True
		}
		opts_str = FakeCmd._process_opts(
			run_opts,
			FakeCmd.opts,
			FakeCmd.opt_key_map,
			FakeCmd.opt_value_map)
		self.assertEqual(opts_str, '-opt1 41.0 --opt3')

	def test_process_opts_with_prefix(self):
		run_opts = {
			'fakecmd_opt1': 41, # should override opt1 below
			'opt1': 45,
			'opt2': False, # intentionally omit arg, overriding default value
			'opt3': True
		}
		opts_str = FakeCmd._process_opts(
			run_opts,
			FakeCmd.opts,
			FakeCmd.opt_key_map,
			FakeCmd.opt_value_map,
			command_name='fakecmd')
		self.assertEqual(opts_str, '-opt1 41.0 --opt3')

	def test_process_opts_with_unsupported(self):
		run_opts = {
			'fakecmd_opt1': 41, # should override opt1 below
			'opt1': 45,
			'opt2': False, # intentionally omit arg, overriding default value
			'opt3': True,
			'opt4': 'test_unsupported'
		}
		opts_str = FakeCmd._process_opts(
			run_opts,
			FakeCmd.opts,
			FakeCmd.opt_key_map,
			FakeCmd.opt_value_map,
			command_name='fakecmd')
		self.assertEqual(opts_str, '-opt1 41.0 --opt3')

	def test_process_opts_with_convert_underscore(self):
		run_opts = {
			'fakecmd_opt1': 41, # should override opt1 below
			'opt1': 45,
			'opt2': False, # intentionally omit arg, overriding default value
			'opt3': True,
			'opt4': 'test_unsupported',
			'opt_with_underscore': 'test'
		}
		opts_str = FakeCmd._process_opts(
			run_opts,
			FakeCmd.opts,
			FakeCmd.opt_key_map,
			FakeCmd.opt_value_map,
			command_name='fakecmd')
		self.assertEqual(opts_str, '-opt1 41.0 --opt3 -opt-with-underscore test')

	def test_get_opt_value(self):
		run_opts = {
			'fakecmd_opt1': 41,
			'opt1': 45
		}
		opt_value = FakeCmd._get_opt_value(
			run_opts,
			opt_name='opt1',
			opt_prefix='fakecmd',
			default=10)
		self.assertEqual(opt_value, 41)

	def test_get_opt_value_false(self):
		run_opts = {
			'fakecmd_opt1': False,
			'opt1': 45
		}
		opt_value = FakeCmd._get_opt_value(
			run_opts,
			opt_name='opt1',
			opt_prefix='fakecmd',
			default=10)
		self.assertEqual(opt_value, False)

	def test_get_opt_value_not_supported(self):
		run_opts = {
			'fakecmd_opt1': False,
			'opt1': 45,
			'opt4': OPT_NOT_SUPPORTED
		}
		opt_value = FakeCmd._get_opt_value(
			run_opts,
			opt_name='opt4',
			opt_prefix='fakecmd',
			default=10)
		self.assertEqual(opt_value, None)

	# def test_httpx_build_cmd_defaults(self):
	# 	if httpx not in TEST_TASKS:
	# 		return
	# 	run_opts = {}
	# 	host = 'test.synology.me'
	# 	cls = httpx(host, **run_opts)
	# 	default_threads = cls.meta_opts[THREADS]['default']
	# 	expected_cmd = f'httpx {DEFAULT_HTTPX_FLAGS} -u {host} -json -rstr {CONFIG.http.response_max_size_bytes} -rsts {CONFIG.http.response_max_size_bytes} -threads {default_threads}'
	# 	self.assertEqual(cls.cmd, expected_cmd)
	# 	self.assertEqual(cls.print_line, False)
	# 	self.assertEqual(cls.print_item, False)

	# def test_httpx_build_cmd_with_opts(self):
	# 	if httpx not in TEST_TASKS:
	# 		return
	# 	run_opts = {
	# 		FOLLOW_REDIRECT: False,
	# 		DELAY: 1,
	# 		RATE_LIMIT: 120,
	# 		THREADS: 10,
	# 		TIMEOUT: 1,
	# 		HEADER: 'Content-Type: application/xml',
	# 		MATCH_CODES: False, # intentionally omit arg, overriding default value
	# 		'filter_codes': '500',
	# 		'filter_size': '23,33'
	# 	}
	# 	host = 'test.synology.me'
	# 	cls = httpx(host, **run_opts)
	# 	expected_cmd = f"httpx {DEFAULT_HTTPX_FLAGS} -u {host} -json -rstr {CONFIG.http.response_max_size_bytes} -rsts {CONFIG.http.response_max_size_bytes} -header 'Content-Type: application/xml' -delay 1s -rate-limit 120 -threads 10 -timeout 1 -filter-code 500 -filter-length 23,33"
	# 	self.assertEqual(cls.cmd, expected_cmd)
	# 	self.assertEqual(cls.print_line, False)
	# 	self.assertEqual(cls.print_item, False)

	# def test_httpx_build_cmd_with_opts_with_prefix(self):
	# 	if httpx not in TEST_TASKS:
	# 		return
	# 	run_opts = {
	# 		FOLLOW_REDIRECT: False,
	# 		DELAY: 1,
	# 		RATE_LIMIT: 120,
	# 		THREADS: 10,
	# 		TIMEOUT: 1,
	# 		HEADER: 'Content-Type: application/xml',
	# 		MATCH_CODES: False, # intentionally omit arg, overriding default value
	# 		'filter_code': '200',
	# 		'filter_length': 50,
	# 		'httpx.filter_codes': '500',    # prefixed option keys should override
	# 		'httpx_filter_size': '23,33' # prefixed option keys should override
	# 	}
	# 	host = 'test.synology.me'
	# 	cls = httpx(host, **run_opts)
	# 	expected_cmd = f"httpx {DEFAULT_HTTPX_FLAGS} -u {host} -json -rstr {CONFIG.http.response_max_size_bytes} -rsts {CONFIG.http.response_max_size_bytes} -header 'Content-Type: application/xml' -delay 1s -rate-limit 120 -threads 10 -timeout 1 -filter-code 500 -filter-length 23,33"
	# 	self.assertEqual(cls.cmd, expected_cmd)
	# 	self.assertEqual(cls.print_line, False)
	# 	self.assertEqual(cls.print_item, False)


class TestCommandRun(unittest.TestCase, CommandOutputTester):

	def setUp(self):
		warnings.simplefilter('ignore', category=ResourceWarning)
		warnings.simplefilter('ignore', category=DeprecationWarning)

	def _valid_fixture(self, cls, fixture):
		if not fixture:
			if len(FIXTURES_TASKS.keys()) == 1: # make test fail.
				raise AssertionError(f'No fixture for {cls.__name__}! Add one to the tests/fixtures directory (must not be an empty file / empty json / empty list).')
			console.print('[dim gold3] skipped (no fixture)[/]')
			return False
		return True

	def test_cmd_converted_schema(self):
		console.print('')

		for cls, fixture in FIXTURES_TASKS.items():
			console.print(f'\t[bold grey35]{cls.__name__} ...[/] ', end='')
			with self.subTest(name=cls.__name__):
				# Validate fixture
				if not self._valid_fixture(cls, fixture):
					continue

				# Run command
				targets = INPUTS_TASKS[cls.input_type]
				with mock_command(cls, targets, META_OPTS, fixture) as runner:
					self._test_runner_output(
						runner,
						expected_output_types=cls.output_types
					)


class TestCommandHooks(unittest.TestCase):

	def test_cmd_hooks(self):
		if httpx not in TEST_TASKS:
			return

		def on_item_pre_convert(self, item):
			item['url'] = 'test_changed_url'
			return item

		def on_item(self, item):
			item.status_code = 500
			return item

		def on_end(self):
			self.results = [{'url': 'test_changed_result'}]

		def on_init(self):
			self.cmd = 'test_changed_cmd_init'
			self.run_opts = {}

		def on_start(self):
			self.cmd = 'test_changed_cmd_start'

		hooks = {
			'on_init': [on_init],
			'on_start': [on_start],
			'on_end': [on_end],
			'on_item_pre_convert': [on_item_pre_convert],
			'on_item': [on_item],
		}
		fixture = load_fixture('httpx_output', FIXTURES_DIR)
		with mock_subprocess_popen([json.dumps(fixture)]):
			input = INPUTS_TASKS[HOST]
			cls = httpx(input, hooks=hooks)
			self.assertEqual(cls.cmd.split(' ')[0], 'test_changed_cmd_init')
			items = cls.run()
			item = items[1]
			self.assertIsInstance(item, Url)
			self.assertEqual(item.status_code, 500)
			self.assertEqual(item.url, 'test_changed_url')
			self.assertEqual(cls.cmd.split(' ')[0], 'test_changed_cmd_start')
			self.assertEqual(cls.results, [{'url': 'test_changed_result'}])

	def test_cmd_failed_hook(self):
		if httpx not in TEST_TASKS:
			return

		def raise_exc(self):
			raise Exception('Test passed')

		hooks = {
			'on_init': [raise_exc]
		}
		fixture = FIXTURES_TASKS[httpx]
		with mock_subprocess_popen([json.dumps(fixture)]):
			with self.assertRaises(Exception, msg='Test passed'):
				input = INPUTS_TASKS[HOST]
				httpx(input, hooks=hooks, raise_on_error=True) 


import logging
import os
import re

import xmltodict

from secator.config import CONFIG
from secator.decorators import task
from secator.definitions import (CONFIDENCE, CVSS_SCORE, DELAY,
								 DESCRIPTION, EXTRA_DATA, FOLLOW_REDIRECT,
								 HEADER, HOST, ID, IP, PROTOCOL, MATCHED_AT, NAME,
								 OPT_NOT_SUPPORTED, OUTPUT_PATH, PORT, PORTS, PROVIDER,
								 PROXY, RATE_LIMIT, REFERENCE, REFERENCES,
								 RETRIES, SCRIPT, SERVICE_NAME, SEVERITY, STATE, TAGS,
								 THREADS, TIMEOUT, TOP_PORTS, USER_AGENT)
from secator.output_types import Exploit, Port, Vulnerability, Info, Error
from secator.tasks._categories import VulnMulti
from secator.utils import debug, traceback_as_string

logger = logging.getLogger(__name__)


@task()
class nmap(VulnMulti):
	"""Network Mapper is a free and open source utility for network discovery and security auditing."""
	cmd = 'nmap'
	input_flag = None
	input_chunk_size = 1
	file_flag = '-iL'
	opt_prefix = '--'
	output_types = [Port, Vulnerability, Exploit]
	opts = {
		PORTS: {'type': str, 'short': 'p', 'help': 'Ports to scan'},
		TOP_PORTS: {'type': int, 'short': 'tp', 'help': 'Top ports to scan [full, 100, 1000]'},
		SCRIPT: {'type': str, 'default': 'vulners', 'help': 'NSE scripts'},
		'skip_host_discovery': {'is_flag': True, 'short': 'Pn', 'default': False, 'help': 'Skip host discovery (no ping)'},
		'version_detection': {'is_flag': True, 'short': 'sV', 'default': False, 'help': 'Version detection'},
		'tcp_syn_stealth': {'is_flag': True, 'short': 'sS', 'default': False, 'help': 'TCP SYN Stealth'},
		'tcp_connect': {'is_flag': True, 'short': 'sT', 'default': False, 'help': 'TCP Connect scan'},
		'udp_scan': {'is_flag': True, 'short': 'sU', 'default': False, 'help': 'UDP scan'},
		'output_path': {'type': str, 'short': 'oX', 'default': None, 'help': 'Output XML file path'},
	}
	opt_key_map = {
		HEADER: OPT_NOT_SUPPORTED,
		DELAY: 'scan-delay',
		FOLLOW_REDIRECT: OPT_NOT_SUPPORTED,
		PROXY: None,  # TODO: supports --proxies but not in TCP mode [https://github.com/nmap/nmap/issues/1098]
		RATE_LIMIT: 'max-rate',
		RETRIES: 'max-retries',
		THREADS: OPT_NOT_SUPPORTED,
		TIMEOUT: 'max-rtt-timeout',
		USER_AGENT: OPT_NOT_SUPPORTED,

		# Nmap opts
		PORTS: '-p',
		'skip_host_discovery': '-Pn',
		'version_detection': '-sV',
		'tcp_connect': '-sT',
		'tcp_syn_stealth': '-sS',
		'udp_scan': '-sU',
		'output_path': '-oX',
	}
	opt_value_map = {
		PORTS: lambda x: ','.join([str(p) for p in x]) if isinstance(x, list) else x
	}
	install_pre = {
		'apt|pacman|brew': ['nmap'],
		'apk': ['nmap', 'nmap-scripts'],
	}
	install_cmd = (
		'sudo git clone --depth 1 --single-branch https://github.com/scipag/vulscan /opt/scipag_vulscan || true && '
		'sudo ln -s /opt/scipag_vulscan /usr/share/nmap/scripts/vulscan || true'
	)
	proxychains = True
	proxychains_flavor = 'proxychains4'
	proxy_socks5 = False
	proxy_http = False
	profile = 'io'

	@staticmethod
	def on_init(self):
		output_path = self.get_opt_value(OUTPUT_PATH)
		if not output_path:
			output_path = f'{self.reports_folder}/.outputs/{self.unique_name}.xml'
		self.output_path = output_path
		self.cmd += f' -oX {self.output_path}'
		tcp_syn_stealth = self.get_opt_value('tcp_syn_stealth')
		tcp_connect = self.get_opt_value('tcp_connect')
		udp_scan = self.get_opt_value('udp_scan')
		if tcp_syn_stealth or udp_scan:
			self.cmd = f'sudo {self.cmd}'
		if tcp_connect and tcp_syn_stealth:
			self._print(
				'Options -sT (SYN stealth scan) and -sS (CONNECT scan) are conflicting. Keeping only -sT.',
				'bold gold3')
			self.cmd = self.cmd.replace('-sT ', '')

	@staticmethod
	def on_cmd_done(self):
		if not os.path.exists(self.output_path):
			yield Error(message=f'Could not find XML results in {self.output_path}')
			return
		yield Info(message=f'XML results saved to {self.output_path}')
		yield from self.xml_to_json()

	def xml_to_json(self):
		results = []
		with open(self.output_path, 'r') as f:
			content = f.read()
			try:
				results = xmltodict.parse(content)  # parse XML to dict
			except Exception as exc:
				yield Error(
					message=f'Cannot parse XML output {self.output_path} to valid JSON.',
					traceback=traceback_as_string(exc)
				)
		yield from nmapData(results)


class nmapData(dict):

	def __iter__(self):
		datas = []
		for host in self._get_hosts():
			hostname = self._get_hostname(host)
			ip = self._get_ip(host)
			for port in self._get_ports(host):
				# Get port number
				port_number = port['@portid']
				if not port_number or not port_number.isdigit():
					continue
				port_number = int(port_number)

				# Get port state
				state = port.get('state', {}).get('@state', '')

				# Get extra data
				extra_data = self._get_extra_data(port)
				service_name = extra_data.get('service_name', '')
				version_exact = extra_data.get('version_exact', False)
				conf = extra_data.get('confidence')

				# Grab CPEs
				cpes = extra_data.get('cpe', [])

				# Get script output
				scripts = self._get_scripts(port)

				# Get port protocol
				protocol = port['@protocol'].lower()

				# Yield port data
				port = {
					PORT: port_number,
					HOST: hostname,
					STATE: state,
					SERVICE_NAME: service_name,
					IP: ip,
					PROTOCOL: protocol,
					EXTRA_DATA: extra_data,
					CONFIDENCE: conf
				}
				yield port

				# Parse each script output to get vulns
				for script in scripts:
					script_id = script['id']
					output = script['output']
					extra_data = {'script': script_id}
					if service_name:
						extra_data['service_name'] = service_name
					funcmap = {
						'vulscan': self._parse_vulscan_output,
						'vulners': self._parse_vulners_output,
					}
					func = funcmap.get(script_id)
					metadata = {
						MATCHED_AT: f'{hostname}:{port_number}',
						IP: ip,
						EXTRA_DATA: extra_data,
					}
					if not func:
						debug(f'Script output parser for "{script_id}" is not supported YET.', sub='cve')
						continue
					for data in func(output, cpes=cpes):
						data.update(metadata)
						confidence = 'low'
						if 'cpe-match' in data[TAGS]:
							confidence = 'high' if version_exact else 'medium'
						data[CONFIDENCE] = confidence
						if (CONFIG.runners.skip_cve_low_confidence and data[CONFIDENCE] == 'low'):
							debug(f'{data[ID]}: ignored (low confidence).', sub='cve')
							continue
						if data in datas:
							continue
						yield data
						datas.append(data)

	#---------------------#
	# XML FILE EXTRACTORS #
	#---------------------#
	def _get_hosts(self):
		hosts = self.get('nmaprun', {}).get('host', {})
		if isinstance(hosts, dict):
			hosts = [hosts]
		return hosts

	def _get_ports(self, host_cfg):
		ports = host_cfg.get('ports', {}).get('port', [])
		if isinstance(ports, dict):
			ports = [ports]
		return ports

	def _get_hostname(self, host_cfg):
		hostnames = host_cfg.get('hostnames', {})
		if hostnames:
			hostnames = hostnames.get('hostname', [])
			if isinstance(hostnames, dict):
				hostnames = [hostnames]
			if hostnames:
				hostname = hostnames[0]['@name']
		else:
			hostname = self._get_address(host_cfg).get('@addr', None)
		return hostname

	def _get_address(self, host_cfg):
		if isinstance(host_cfg.get('address', {}), list):
			addresses = host_cfg.get('address', {})
			for address in addresses:
				if address.get('@addrtype') == "ipv4":
					return address
		return host_cfg.get('address', {})

	def _get_ip(self, host_cfg):
		return self._get_address(host_cfg).get('@addr', None)

	def _get_extra_data(self, port_cfg):
		extra_data = {
			k.lstrip('@'): v
			for k, v in port_cfg.get('service', {}).items()
		}

		# Strip product / version strings
		if 'product' in extra_data:
			extra_data['product'] = extra_data['product'].lower()

		# Get version and post-process it
		version = None
		if 'version' in extra_data:
			vsplit = extra_data['version'].split(' ')
			version_exact = True
			os = None
			if len(vsplit) == 3:
				version, os, extra_version = tuple(vsplit)
				if os == 'or' and extra_version == 'later':
					version_exact = False
					os = None
				version = f'{version}-{extra_version}'
			elif len(vsplit) == 2:
				version, os = tuple(vsplit)
			elif len(vsplit) == 1:
				version = vsplit[0]
			else:
				version = extra_data['version']
			if os:
				extra_data['os'] = os
			if version:
				extra_data['version'] = version
			extra_data['version_exact'] = version_exact

		# Grap service name
		product = extra_data.get('product', None) or extra_data.get('name', None)
		if product:
			service_name = product
			if version:
				service_name += f'/{version}'
			extra_data['service_name'] = service_name

		# Grab CPEs
		cpes = extra_data.get('cpe', [])
		if not isinstance(cpes, list):
			cpes = [cpes]
			extra_data['cpe'] = cpes
		debug(f'Found CPEs: {",".join(cpes)}', sub='cve')

		# Grab confidence
		conf = int(extra_data.get('conf', 0))
		if conf > 7:
			confidence = 'high'
		elif conf > 4:
			confidence = 'medium'
		else:
			confidence = 'low'
		extra_data['confidence'] = confidence

		# Build custom CPE
		if product and version:
			vsplit = version.split('-')
			version_cpe = vsplit[0] if not version_exact else version
			cpe = VulnMulti.create_cpe_string(product, version_cpe)
			if cpe not in cpes:
				cpes.append(cpe)
				debug(f'Added new CPE from identified product and version: {cpe}', sub='cve')

		return extra_data

	def _get_scripts(self, port_cfg):
		scripts = port_cfg.get('script', [])
		if isinstance(scripts, dict):
			scripts = [scripts]
		scripts = [
			{k.lstrip('@'): v for k, v in script.items()}
			for script in scripts
		]
		return scripts

	#--------------#
	# VULN PARSERS #
	#--------------#
	def _parse_vulscan_output(self, out, cpes=[]):
		"""Parse nmap vulscan script output.

		Args:
			out (str): Vulscan script output.

		Returns:
			list: List of Vulnerability dicts.
		"""
		provider_name = ''
		for line in out.splitlines():
			if not line:
				continue
			line = line.strip()
			if not line.startswith('[') and line != 'No findings':  # provider line
				provider_name, _ = tuple(line.split(' - '))
				continue
			reg = r'\[([ A-Za-z0-9_@./#&+-]*)\] (.*)'
			matches = re.match(reg, line)
			if not matches:
				continue
			vuln_id, vuln_title = matches.groups()
			vuln = {
				ID: vuln_id,
				NAME: vuln_id,
				DESCRIPTION: vuln_title,
				PROVIDER: provider_name,
				TAGS: [vuln_id, provider_name]
			}
			if provider_name == 'MITRE CVE':
				data = VulnMulti.lookup_cve(vuln['id'], *cpes)
				if data:
					vuln.update(data)
				yield vuln
			else:
				debug(f'Vulscan provider {provider_name} is not supported YET.', sub='cve.provider', verbose=True)
				continue

	def _parse_vulners_output(self, out, **kwargs):
		cpes = kwargs.get('cpes', [])
		provider_name = 'vulners'
		for line in out.splitlines():
			if not line:
				continue
			line = line.strip()
			if line.startswith('cpe:'):
				cpes.append(line.rstrip(':'))
				continue
			elems = tuple(line.split('\t'))

			if len(elems) == 4:  # exploit
				# TODO: Implement exploit processing
				exploit_id, cvss_score, reference_url, _ = elems
				name = exploit_id
				# edb_id = name.split(':')[-1] if 'EDB-ID' in name else None
				exploit = {
					ID: exploit_id,
					NAME: name,
					PROVIDER: provider_name,
					REFERENCE: reference_url,
					TAGS: [exploit_id, provider_name],
					CVSS_SCORE: cvss_score,
					CONFIDENCE: 'low',
					'_type': 'exploit',
				}
				# TODO: lookup exploit in ExploitDB to find related CVEs
				# if edb_id:
				# 	print(edb_id)
				# 	exploit_data = VulnMulti.lookup_exploitdb(edb_id)
				vuln = VulnMulti.lookup_cve_from_vulners_exploit(exploit_id, *cpes)
				if vuln:
					yield vuln
					exploit[TAGS].extend(vuln[TAGS])
					exploit[CONFIDENCE] = vuln[CONFIDENCE]
				yield exploit

			elif len(elems) == 3:  # vuln
				vuln = {}
				vuln_id, vuln_cvss, reference_url = tuple(line.split('\t'))
				vuln_cvss = float(vuln_cvss)
				vuln_id = vuln_id.split(':')[-1]
				vuln_type = vuln_id.split('-')[0]
				vuln = {
					ID: vuln_id,
					NAME: vuln_id,
					PROVIDER: provider_name,
					CVSS_SCORE: vuln_cvss,
					SEVERITY: VulnMulti.cvss_to_severity(vuln_cvss),
					REFERENCES: [reference_url],
					TAGS: [vuln_id, provider_name],
					CONFIDENCE: 'low'
				}
				if vuln_type == 'CVE' or vuln_type == 'PRION:CVE':
					data = VulnMulti.lookup_cve(vuln_id, *cpes)
					if data:
						vuln.update(data)
					yield vuln
				else:
					debug(f'Vulners parser for "{vuln_type}" is not implemented YET.', sub='cve')
			else:
				debug(f'Unrecognized vulners output: {elems}', sub='cve')



import queue
import os
import unittest
import warnings

from time import sleep
from threading import Thread

from celery import chain, chord

from secator.celery import app, forward_results
from secator.config import CONFIG
from secator.utils_test import TEST_TASKS, load_fixture
from secator.runners import Command
from secator.output_types import Url
from tests.integration.inputs import INPUTS_SCANS


INTEGRATION_DIR = os.path.dirname(os.path.abspath(__file__))
OPTS = {
	'ffuf.filter_size': '3748,3106',
	'ffuf.depth': 1,
	'ffuf.follow_redirect': True,
	'ffuf.wordlist': load_fixture('wordlist', INTEGRATION_DIR, only_path=True),
}
URL_TARGETS = INPUTS_SCANS['url']
URL_TARGETS_RESULTS_COUNT = [12, 1]
HOST_TARGETS = INPUTS_SCANS['host']


class TestCelery(unittest.TestCase):

	@classmethod
	def setUpClass(cls):
		warnings.simplefilter('ignore', category=ResourceWarning)
		warnings.simplefilter('ignore', category=DeprecationWarning)
		Command.execute(
			f'sh {INTEGRATION_DIR}/setup.sh',
			quiet=True,
			cwd=INTEGRATION_DIR
		)
		cls.queue = queue.Queue()
		cls.cmd = Command.execute('secator worker', quiet=True, run=False)
		cls.thread = Thread(target=cls.cmd.run)
		cls.thread.start()
		sleep(5)

	@classmethod
	def tearDownClass(cls) -> None:
		cls.cmd.stop_process()
		cls.thread.join()
		Command.execute(
			f'sh {INTEGRATION_DIR}/teardown.sh',
			quiet=True,
			cwd=INTEGRATION_DIR
		)

	def test_httpx_chain(self):
		from secator.tasks import httpx
		if httpx not in TEST_TASKS:
			return
		sigs = [forward_results.si([])] + [httpx.s(target) for target in URL_TARGETS]
		workflow = chain(*sigs)
		result = workflow.apply()
		results = result.get()
		urls = [r.url for r in results if r._type == 'url']
		targets = [r.name for r in results if r._type == 'target']
		self.assertEqual(len(urls), len(URL_TARGETS))
		self.assertEqual(len(targets), len(URL_TARGETS))

	def test_httpx_chain_prior_results(self):
		from secator.tasks import httpx
		if httpx not in TEST_TASKS:
			return

		existing_results = [Url(**{
			"url": "https://example.synology.me",
			"method": "GET",
			"status_code": 200,
			"words": 438,
			"lines": 136,
			"content_type":
			"text/html",
			"content_length": 11577,
			"host": "82.66.157.114",
			"time": 0.16246860100000002,
			"_source": "httpx",
			"_type": "url"
		})]
		targets = INPUTS_SCANS['url']
		sigs = [forward_results.s(existing_results)] + [httpx.s(target) for target in URL_TARGETS]
		workflow = chain(*sigs)
		result = workflow.apply()
		results = result.get()
		urls = [r.url for r in results if r._type == 'url']
		targets = [r.name for r in results if r._type == 'target']
		self.assertEqual(len(urls), len(URL_TARGETS) + 1)
		self.assertEqual(len(targets), len(URL_TARGETS))
		self.assertIn(existing_results[0], results)

	def test_httpx_chord(self):
		from secator.tasks import httpx
		if httpx not in TEST_TASKS:
			return

		existing_results = [Url(**{
			"url": "https://example.synology.me",
			"method": "GET",
			"status_code": 200,
			"words": 438,
			"lines": 136,
			"content_type":
			"text/html",
			"content_length": 11577,
			"host": "82.66.157.114",
			"time": 0.16246860100000002,
			"_source": "httpx",
			"_type": "url"
		})]

		sigs = []
		for target in URL_TARGETS:
			sig = httpx().s(target)
			sigs.append(sig)
		workflow = chain(
			forward_results.s(existing_results),
			sigs[0],
			chord((
				sigs[1],
				sigs[0],
			), forward_results.s()),
			sigs[1],
			chord((
				sigs[0],
				sigs[1],
			), forward_results.s())
		)
		result = workflow.apply()
		results = result.get()
		urls = [r.url for r in results if r._type == 'url']
		targets = [r.name for r in results if r._type == 'target']
		self.assertIn(existing_results[0], results)
		self.assertEqual(len(targets), len(targets))
		self.assertEqual(len(urls), len(targets) + 1)

	def test_httpx_chunk(self):
		from secator.tasks import httpx
		if httpx not in TEST_TASKS:
			return

		size = CONFIG.runners.input_chunk_size + 1
		targets = [URL_TARGETS[0]] * size
		result = httpx.delay(targets)
		results = result.get()
		urls = [r.url for r in results if r._type == 'url']
		infos = [r.message for r in results if r._type == 'info']
		self.assertEqual(len(urls), 2)  # same URL, but twice because 2 chunks and same input
		# self.assertEqual(len(infos), 2) # one chunk message for each chunk
		# for message in infos:
			# self.assertIn('Celery chunked task created', message)

	def test_nmap_chain(self):
		from secator.tasks import nmap
		if nmap not in TEST_TASKS:
			return

		workflow = chain(
			forward_results.s([]),
			chord((
				nmap.s(URL_TARGETS)
			), forward_results.s()),
		)
		result = workflow.apply()
		results = result.get()
		targets = [r.name for r in results if r._type == 'target']
		self.assertEqual(len(targets), len(URL_TARGETS))

	def test_ffuf_chunked(self):
		from secator.tasks import ffuf
		if ffuf not in TEST_TASKS:
			return

		targets = [t + '/FUZZ' for t in URL_TARGETS]
		workflow = chain(
			forward_results.s([]),
			chord((
				ffuf.s(targets, **OPTS)
			), forward_results.s()),
		)
		result = workflow.apply()
		results = result.get()
		targets = [r.name for r in results if r._type == 'target']
		urls = [r.url for r in results if r._type == 'url']
		self.assertEqual(len(targets), len(URL_TARGETS))
		self.assertEqual(len(urls), sum(URL_TARGETS_RESULTS_COUNT))



import logging
import os
import unittest
import warnings
from time import sleep

from secator.template import TemplateLoader
from secator.runners import Task
from secator.output_types import Port, Url
from secator.definitions import DEBUG
from secator.runners import Command, Workflow
from secator.utils import setup_logging, merge_opts
from secator.utils_test import TEST_WORKFLOWS, ALL_WORKFLOWS, CommandOutputTester, load_fixture
from tests.integration.inputs import INPUTS_WORKFLOWS
from tests.integration.outputs import OUTPUTS_WORKFLOWS

INTEGRATION_DIR = os.path.dirname(os.path.abspath(__file__))
level = logging.DEBUG if DEBUG > 0 else logging.INFO
setup_logging(level)


def hook_workflow_init(self):
	self.context['workflow_id'] = 1


def hook_task_init(self):
	self.context['task_id'] = 1


def hook_item(self, item):
	print(item.toDict())
	return item


class TestWorkflows(unittest.TestCase, CommandOutputTester):

	@classmethod
	def setUpClass(cls):
		warnings.simplefilter('ignore', category=ResourceWarning)
		warnings.simplefilter('ignore', category=DeprecationWarning)
		Command.execute(
			f'sh {INTEGRATION_DIR}/setup.sh',
			quiet=True,
			cwd=INTEGRATION_DIR
		)
		sleep(15)

	@classmethod
	def tearDownClass(self):
		Command.execute(
			f'sh {INTEGRATION_DIR}/teardown.sh',
			quiet=True,
			cwd=INTEGRATION_DIR
		)

	def test_default_workflows(self):
		fmt_opts = {
			'print_item': DEBUG > 1,
			'print_line': DEBUG > 2,
			'table': DEBUG > 1,
			'output': 'table' if DEBUG > 1 else ''
		}
		opts = {
			'ffuf.filter_size': 1987,
			'feroxbuster.filter_size': 1987,
			'dirsearch.filter_size': 1987,
			'follow_redirect': True,
			'match_codes': '200',
			'httpx.match_codes': False,
			'httpx.filter_size': False,
			'nuclei.retries': 5,
			'nuclei.timeout': 15,
			'rate_limit': 1000,
			'wordlist': load_fixture('wordlist', INTEGRATION_DIR, only_path=True),
			'timeout': 7,
			'depth': 2,
			'ports': '9999,3000,8080'
		}
		opts = merge_opts(opts, fmt_opts)

		for conf in TEST_WORKFLOWS:
			with self.subTest(name=conf.name):
				inputs = INPUTS_WORKFLOWS.get(conf.name, [])
				outputs = OUTPUTS_WORKFLOWS.get(conf.name, [])
				workflow = Workflow(conf, inputs=inputs, run_opts=opts)
				self._test_runner_output(
					workflow,
					expected_results=outputs)

	def test_inline_workflow(self):
		# Ignore if TEST_WORKFLOWS are defined
		if TEST_WORKFLOWS != ALL_WORKFLOWS:
			return

		# Expected results / context
		expected_results = [
			Port(port=9999, host='localhost', ip='127.0.0.1', state='open', service_name='fake', _source='unknown'),
			Port(port=3000, host='localhost', ip='127.0.0.1', state='open', _source='naabu'),
			Port(port=8080, host='localhost', ip='127.0.0.1', state='open', _source='naabu'),
			Url(url='http://localhost:3000', host='127.0.0.1', status_code=200, title='OWASP Juice Shop', content_type='text/html', _source='httpx'),
			Url(url='http://localhost:8080', host='127.0.0.1', status_code=400, title='', content_type='application/json', _source='httpx'),
		]
		expected_context = {
			'task_id': 1,
			'workflow_id': 1
		}

		# Create ad-hoc workflow
		conf = {
			'name': 'my_workflow',
			'type': 'workflow',
			'description': 'Test workflow',
			'tasks': {
				'naabu': {},
				'httpx': {
					'targets_': {'type': 'port', 'field': '{host}:{port}'}
				}
			}
		}
		config = TemplateLoader(conf)
		workflow = Workflow(
			config,
			inputs=['localhost'],
			results=[
				Port(port=9999, host='localhost', ip='127.0.0.1', state='open', service_name='fake', _source='unknown', _context=expected_context)
			],
			hooks = {
				Workflow: {
					'on_init': [hook_workflow_init],
				},
				Task: {
					'on_init': [hook_task_init],
					'on_item': [hook_item],
				}
			}
		)
		uuids = []
		results = []

		# Verify no duplicates and context added from hook is present in output
		for result in workflow:
			self.assertEqual(result._context, {**result._context, **expected_context})
			self.assertNotIn(result._uuid, uuids)
			uuids.append(result._uuid)
			results.append(result)

		for res in expected_results:
			self.assertIn(res, workflow.results)



import json
import unittest

from pathlib import Path

from secator.utils_test import load_fixture, FIXTURES_DIR
from secator.tasks._categories import Vuln
from secator.config import CONFIG


class TestSerializers(unittest.TestCase):

	def test_lookup_cve_circle(self):
		fixture = json.dumps(load_fixture('cve_circle_output', FIXTURES_DIR), sort_keys=True)
		cve_path = f'{CONFIG.dirs.data}/cves/CVE-2023-5568.json'
		if Path(cve_path).exists():
			Path(cve_path).unlink()  # make sure we don't use cache data
		actual = json.dumps(Vuln.lookup_cve_from_cve_circle('CVE-2023-5568'), sort_keys=True)
		self.assertEqual(actual, fixture)



import os

from secator.decorators import task
from secator.definitions import (DELAY, DEPTH, FILTER_CODES, FILTER_REGEX, FILTER_SIZE, FILTER_WORDS, FOLLOW_REDIRECT,
								 HEADER, MATCH_CODES, MATCH_REGEX, MATCH_SIZE, MATCH_WORDS, METHOD, OPT_NOT_SUPPORTED,
								 PROXY, RATE_LIMIT, RETRIES, THREADS, TIMEOUT, URL, USER_AGENT)
from secator.config import CONFIG
from secator.output_types import Url, Subdomain
from secator.serializers import JSONSerializer
from secator.tasks._categories import Http
from secator.utils import (sanitize_url, extract_domain_info, extract_subdomains_from_fqdn)


@task()
class httpx(Http):
	"""Fast and multi-purpose HTTP toolkit."""
	cmd = 'httpx'
	file_flag = '-l'
	input_flag = '-u'
	json_flag = '-json'
	opts = {
		# 'silent': {'is_flag': True, 'default': False, 'help': 'Silent mode'},
		# 'irr': {'is_flag': True, 'default': False, 'help': 'Include http request / response'},
		'fep': {'is_flag': True, 'default': False, 'help': 'Error Page Classifier and Filtering'},
		'favicon': {'is_flag': True, 'default': False, 'help': 'Favicon hash'},
		'jarm': {'is_flag': True, 'default': False, 'help': 'Jarm fingerprint'},
		'asn': {'is_flag': True, 'default': False, 'help': 'ASN detection'},
		'cdn': {'is_flag': True, 'default': False, 'help': 'CDN detection'},
		'debug_resp': {'is_flag': True, 'default': False, 'help': 'Debug response'},
		'vhost': {'is_flag': True, 'default': False, 'help': 'Probe and display server supporting VHOST'},
		'store_responses': {'is_flag': True, 'short': 'sr', 'default': CONFIG.http.store_responses, 'help': 'Save HTTP responses'},  # noqa: E501
		'screenshot': {'is_flag': True, 'short': 'ss', 'default': False, 'help': 'Screenshot response'},
		'system_chrome': {'is_flag': True, 'default': False, 'help': 'Use local installed Chrome for screenshot'},
		'headless_options': {'is_flag': False, 'short': 'ho', 'default': None, 'help': 'Headless Chrome additional options'},
		'follow_host_redirects': {'is_flag': True, 'short': 'fhr', 'default': None, 'help': 'Follow redirects on the same host'},  # noqa: E501
		'tech_detect': {'is_flag': True, 'short': 'td', 'default': True, 'help': 'Tech detection'},
		'tls_grab': {'is_flag': True, 'short': 'tlsg', 'default': False, 'help': 'Grab some informations from the tls certificate'},  # noqa: E501
		'rstr': {'type': int, 'default': CONFIG.http.response_max_size_bytes, 'help': 'Max body size to read (bytes)'},
		'rsts': {'type': int, 'default': CONFIG.http.response_max_size_bytes, 'help': 'Max body size to save (bytes)'}
	}
	opt_key_map = {
		HEADER: 'header',
		DELAY: 'delay',
		DEPTH: OPT_NOT_SUPPORTED,
		FILTER_CODES: 'filter-code',
		FILTER_REGEX: 'filter-regex',
		FILTER_SIZE: 'filter-length',
		FILTER_WORDS: 'filter-word-count',
		FOLLOW_REDIRECT: 'follow-redirects',
		MATCH_CODES: 'match-code',
		MATCH_REGEX: 'match-regex',
		MATCH_SIZE: 'match-length',
		MATCH_WORDS: 'match-word-count',
		METHOD: 'x',
		PROXY: 'proxy',
		RATE_LIMIT: 'rate-limit',
		RETRIES: 'retries',
		THREADS: 'threads',
		TIMEOUT: 'timeout',
		USER_AGENT: OPT_NOT_SUPPORTED,
		'store_responses': 'sr',
	}
	opt_value_map = {
		DELAY: lambda x: str(x) + 's' if x else None,
	}
	item_loaders = [JSONSerializer()]
	output_types = [Url, Subdomain]
	install_pre = {
		'apk': ['chromium']
	}
	install_cmd = 'go install -v github.com/projectdiscovery/httpx/cmd/httpx@latest'
	install_github_handle = 'projectdiscovery/httpx'
	proxychains = False
	proxy_socks5 = True
	proxy_http = True
	profile = 'io'

	@staticmethod
	def on_init(self):
		debug_resp = self.get_opt_value('debug_resp')
		if debug_resp:
			self.cmd = self.cmd.replace('-silent', '')
		screenshot = self.get_opt_value('screenshot')
		store_responses = self.get_opt_value('store_responses')
		if store_responses or screenshot:
			self.cmd += f' -srd {self.reports_folder}/.outputs'
		if screenshot:
			self.cmd += ' -esb -ehb'
		self.domains = []

	@staticmethod
	def on_json_loaded(self, item):
		item = self._preprocess_url(item)
		yield item
		tls = item.get('tls', None)
		if tls:
			subject_cn = tls.get('subject_cn', None)
			subject_an = tls.get('subject_an', [])
			cert_domains = subject_an
			if subject_cn:
				cert_domains.append(subject_cn)
			for cert_domain in cert_domains:
				subdomain = self._create_subdomain_from_tls_cert(cert_domain, item['url'])
				if subdomain:
					yield subdomain

	@staticmethod
	def on_end(self):
		store_responses = self.get_opt_value('store_responses')
		response_dir = f'{self.reports_folder}/.outputs'
		if store_responses:
			index_rpath = f'{response_dir}/response/index.txt'
			index_spath = f'{response_dir}/screenshot/index_screenshot.txt'
			index_spath2 = f'{response_dir}/screenshot/screenshot.html'
			if os.path.exists(index_rpath):
				os.remove(index_rpath)
			if os.path.exists(index_spath):
				os.remove(index_spath)
			if os.path.exists(index_spath2):
				os.remove(index_spath2)

	def _preprocess_url(self, item):
		"""Replace time string by float, sanitize URL, get final redirect URL."""
		for k, v in item.items():
			if k == 'time':
				response_time = float(''.join(ch for ch in v if not ch.isalpha()))
				if v[-2:] == 'ms':
					response_time = response_time / 1000
				item[k] = response_time
			elif k == URL:
				item[k] = sanitize_url(v)
		item[URL] = item.get('final_url') or item[URL]
		return item

	def _create_subdomain_from_tls_cert(self, domain, url):
		"""Extract subdomains from TLS certificate."""
		if domain.startswith('*.'):
			domain = domain.lstrip('*.')
		if domain in self.domains:
			return None
		url_domain = extract_domain_info(url)
		url_domains = extract_subdomains_from_fqdn(url_domain.fqdn, url_domain.domain, url_domain.suffix)
		if not url_domain or domain not in url_domains:
			return None
		self.domains.append(domain)
		return Subdomain(
			host=domain,
			domain=extract_domain_info(domain, domain_only=True)
		)



__all__ = [
    'JSONSerializer',
    'RegexSerializer',
    'DataclassEncoder',
]
from secator.serializers.json import JSONSerializer
from secator.serializers.regex import RegexSerializer
from secator.serializers.dataclass import DataclassEncoder



import time
from dataclasses import dataclass, field
from typing import List

from secator.definitions import (CONFIDENCE, CVSS_SCORE, EXTRA_DATA, ID,
								 MATCHED_AT, NAME, REFERENCE, SEVERITY, TAGS)
from secator.output_types import OutputType
from secator.utils import rich_to_ansi, rich_escape as _s


@dataclass
class Vulnerability(OutputType):
	name: str
	provider: str = ''
	id: str = ''
	matched_at: str = ''
	ip: str = field(default='', compare=False)
	confidence: str = 'low'
	severity: str = 'unknown'
	cvss_score: float = 0
	tags: List[str] = field(default_factory=list)
	extra_data: dict = field(default_factory=dict, compare=False)
	description: str = field(default='', compare=False)
	references: List[str] = field(default_factory=list, compare=False)
	reference: str = field(default='', compare=False)
	confidence_nb: int = 0
	severity_nb: int = 0
	_source: str = field(default='', repr=True)
	_type: str = field(default='vulnerability', repr=True)
	_timestamp: int = field(default_factory=lambda: time.time(), compare=False)
	_uuid: str = field(default='', repr=True, compare=False)
	_context: dict = field(default_factory=dict, repr=True, compare=False)
	_tagged: bool = field(default=False, repr=True, compare=False)
	_duplicate: bool = field(default=False, repr=True, compare=False)
	_related: list = field(default_factory=list, compare=False)

	_table_fields = [
		MATCHED_AT,
		SEVERITY,
		CONFIDENCE,
		NAME,
		ID,
		CVSS_SCORE,
		TAGS,
		EXTRA_DATA,
		REFERENCE
	]
	_sort_by = ('confidence_nb', 'severity_nb', 'matched_at', 'cvss_score')

	def __post_init__(self):
		super().__post_init__()
		severity_map = {
			'critical': 0,
			'high': 1,
			'medium': 2,
			'low': 3,
			'info': 4,
			'unknown': 5,
			None: 6
		}
		self.severity = self.severity.lower()  # normalize severity
		self.severity_nb = severity_map.get(self.severity, 6)
		self.confidence_nb = severity_map[self.confidence]
		if len(self.references) > 0:
			self.reference = self.references[0]

	def __repr__(self):
		data = self.extra_data
		if 'data' in data and isinstance(data['data'], list):
			data = ','.join(data['data'])
		elif isinstance(data, dict):
			data = ', '.join([f'{k}:{v}' for k, v in data.items()])
		data = _s(data)
		tags = self.tags
		colors = {
			'critical': 'bold red',
			'high': 'red',
			'medium': 'yellow',
			'low': 'green',
			'info': 'magenta',
			'unknown': 'dim magenta'
		}
		c = colors.get(self.severity, 'dim magenta')
		s = rf'🚨 \[[green]{_s(self.name)} [link={_s(self.reference)}]🡕[/link][/]] \[[{c}]{self.severity}[/]] {_s(self.matched_at)}'  # noqa: E501
		if tags:
			tags_str = ','.join(tags)
			s += rf' \[[cyan]{_s(tags_str)}[/]]'
		if data:
			s += rf' \[[yellow]{str(data)}[/]]'
		if self.confidence == 'low':
			s = f'[dim]{s}[/]'
		return rich_to_ansi(s)

	def __str__(self):
		return self.matched_at + ' -> ' + self.name



import logging
import re
from dataclasses import _MISSING_TYPE, dataclass, fields
from secator.definitions import DEBUG
from secator.rich import console

logger = logging.getLogger(__name__)


@dataclass
class OutputType:
	_table_fields = []
	_sort_by = ()

	def __gt__(self, other):
		if not self.__eq__(other):
			return False

		# Point-based system based on number of non-empty extra-data present.
		# In this configuration, a > b if a == b AND a has more non-empty fields than b
		# extra_fields = [f for f in fields(self) if not f.compare]
		# points1 = 0
		# points2 = 0
		# for field in extra_fields:
		# 	v1 = getattr(self, field.name)
		# 	v2 = getattr(other, field.name)
		# 	if v1 and not v2:
		# 		points1 += 1
		# 	elif v2 and not v1:
		# 		points2 += 1
		# if points1 > points2:
		# 	return True

		# Timestamp-based system: return newest object
		return self._timestamp > other._timestamp

	def __ge__(self, other):
		return self == other

	def __lt__(self, other):
		return other > self

	def __le__(self, other):
		return self == other

	def __post_init__(self):
		"""Initialize default fields to their proper types."""
		for field in fields(self):
			default_factory = field.default_factory
			default = field.default
			if getattr(self, field.name) is None:
				if not isinstance(default, _MISSING_TYPE):
					setattr(self, field.name, field.default)
				elif not isinstance(default_factory, _MISSING_TYPE):
					setattr(self, field.name, default_factory())

	@classmethod
	def load(cls, item, output_map={}):
		new_item = {}

		# Check for explicit _type keys
		_type = item.get('_type')
		if _type and _type != cls.get_name():
			raise TypeError(f'Item has different _type set: {_type}')

		for field in fields(cls):
			key = field.name
			if key in output_map:
				mapped_key = output_map[key]
				if callable(mapped_key):
					try:
						mapped_val = mapped_key(item)
					except Exception as e:
						mapped_val = None
						if DEBUG > 1:
							console.print_exception(show_locals=True)
						raise TypeError(
							f'Fail to transform value for "{key}" using output_map function. Exception: '
							f'{type(e).__name__}: {str(e)}')
				else:
					mapped_val = item.get(mapped_key)
				new_item[key] = mapped_val
			elif key in item:
				new_item[key] = item[key]

		# All values None, raise an error
		if all(val is None for val in new_item.values()):
			raise TypeError(f'Item does not match {cls} schema')

		new_item['_type'] = cls.get_name()
		return cls(**new_item)

	@classmethod
	def get_name(cls):
		return re.sub(r'(?<!^)(?=[A-Z])', '_', cls.__name__).lower()

	@classmethod
	def keys(cls):
		return [f.name for f in fields(cls)]

	def toDict(self, exclude=[]):
		data = self.__dict__.copy()
		if exclude:
			return {k: v for k, v in data.items() if k not in exclude}
		return data



from secator.decorators import task
from secator.definitions import OPT_PIPE_INPUT, OPT_NOT_SUPPORTED, URL
from secator.output_types import Tag
from secator.tasks._categories import Tagger


@task()
class gf(Tagger):
	"""Wrapper around grep, to help you grep for things."""
	cmd = 'gf'
	file_flag = OPT_PIPE_INPUT
	input_flag = OPT_PIPE_INPUT
	version_flag = OPT_NOT_SUPPORTED
	opts = {
		'pattern': {'type': str, 'help': 'Pattern names to match against (comma-delimited)', 'required': True}
	}
	opt_key_map = {
		'pattern': ''
	}
	input_type = URL
	install_cmd = (
		'go install -v github.com/tomnomnom/gf@latest && '
		'git clone https://github.com/1ndianl33t/Gf-Patterns $HOME/.gf || true'
	)
	output_types = [Tag]

	@staticmethod
	def item_loader(self, line):
		yield {'match': line, 'name': self.get_opt_value('pattern').rstrip() + ' pattern'}  # noqa: E731,E501

	@staticmethod
	def on_item(self, item):
		item.extra_data = {'source': 'url'}
		return item



from contextlib import nullcontext
from time import sleep

import kombu
import kombu.exceptions

from celery.result import AsyncResult, GroupResult
from greenlet import GreenletExit
from rich.panel import Panel
from rich.padding import Padding

from rich.progress import Progress as RichProgress, SpinnerColumn, TextColumn, TimeElapsedColumn
from secator.config import CONFIG
from secator.definitions import STATE_COLORS
from secator.output_types import Error
from secator.rich import console
from secator.utils import debug, traceback_as_string


class CeleryData(object):
	"""Utility to simplify tracking a Celery task and all of its subtasks."""

	def iter_results(
			result,
			ids_map={},
			description=True,
			refresh_interval=CONFIG.runners.poll_frequency,
			print_remote_info=True,
			print_remote_title='Results'
		):
		"""Generator to get results from Celery task.

		Args:
			result (Union[AsyncResult, GroupResult]): Celery result.
			description (bool): Whether to show task description.
			refresh_interval (int): Refresh interval.
			print_remote_info (bool): Whether to display live results.
			print_remote_title (str): Title for the progress panel.

		Yields:
			dict: Subtasks state and results.
		"""
		# Display live results if print_remote_info is set
		if print_remote_info:
			class PanelProgress(RichProgress):
				def get_renderables(self):
					yield Padding(Panel(
						self.make_tasks_table(self.tasks),
						title=print_remote_title,
						border_style='bold gold3',
						expand=False,
						highlight=True), pad=(2, 0, 0, 0))
			progress = PanelProgress(
				SpinnerColumn('dots'),
				TextColumn('{task.fields[descr]}  ') if description else '',
				TextColumn('[bold cyan]{task.fields[full_name]}[/]'),
				TextColumn('{task.fields[state]:<20}'),
				TimeElapsedColumn(),
				TextColumn('{task.fields[count]}'),
				TextColumn('{task.fields[progress]}%'),
				# TextColumn('\[[bold magenta]{task.fields[id]:<30}[/]]'),  # noqa: W605
				refresh_per_second=1,
				transient=False,
				console=console,
				# redirect_stderr=True,
				# redirect_stdout=False
			)
		else:
			progress = nullcontext()

		with progress:

			# Make initial progress
			if print_remote_info:
				progress_cache = CeleryData.init_progress(progress, ids_map)

			# Get live results and print progress
			for data in CeleryData.poll(result, ids_map, refresh_interval):
				yield from data['results']

				if print_remote_info:
					task_id = data['id']
					progress_id = progress_cache[task_id]
					CeleryData.update_progress(progress, progress_id, data)

			# Update all tasks to 100 %
			if print_remote_info:
				for progress_id in progress_cache.values():
					progress.update(progress_id, advance=100)

	@staticmethod
	def init_progress(progress, ids_map):
		cache = {}
		for task_id, data in ids_map.items():
			pdata = data.copy()
			state = data['state']
			pdata['state'] = f'[{STATE_COLORS[state]}]{state}[/]'
			id = progress.add_task('', advance=0, **pdata)
			cache[task_id] = id
		return cache

	@staticmethod
	def update_progress(progress, progress_id, data):
		"""Update rich progress with fresh data."""
		pdata = data.copy()
		state = data['state']
		pdata['state'] = f'[{STATE_COLORS[state]}]{state}[/]'
		pdata = {k: v for k, v in pdata.items() if v}
		progress.update(progress_id, **pdata)

	@staticmethod
	def poll(result, ids_map, refresh_interval):
		"""Poll Celery subtasks results in real-time. Fetch task metadata and partial results from each task that runs.

		Yields:
			dict: Subtasks state and results.
		"""
		while True:
			try:
				yield from CeleryData.get_all_data(result, ids_map)
				if result.ready():
					debug('result is ready', sub='celery.poll', id=result.id)
					yield from CeleryData.get_all_data(result, ids_map)
					break
			except (KeyboardInterrupt, GreenletExit):
				debug('encounted KeyboardInterrupt or GreenletExit', sub='celery.poll')
				raise
			except Exception as e:
				error = Error.from_exception(e)
				debug(repr(error), sub='celery.poll')
				pass
			finally:
				sleep(refresh_interval)

	@staticmethod
	def get_all_data(result, ids_map):
		"""Get Celery results from main result object, AND all subtasks results.

		Yields:
			dict: Subtasks state and results.
		"""
		task_ids = list(ids_map.keys())
		for task_id in task_ids:
			data = CeleryData.get_task_data(task_id, ids_map)
			if not data:
				continue
			debug(
				'POLL',
				sub='celery.poll',
				id=data['id'],
				obj={data['full_name']: data['state'], 'count': data['count']},
				verbose=True
			)
			yield data

		# Calculate and yield parent task progress
		# if not datas:
		# 	return
		# total = len(datas)
		# count_finished = sum([i['ready'] for i in datas if i])
		# percent = int(count_finished * 100 / total) if total > 0 else 0
		# parent_id = [c for c in ids_map.values() if c['full_name'] == datas[-1]]
		# data['progress'] = percent
		# yield data

	@staticmethod
	def get_task_data(task_id, ids_map):
		"""Get task info.

		Args:
			task_id (str): Celery task id.

		Returns:
			dict: Task info (id, name, state, results, chunk_info, count, error, ready).
		"""

		# Get task data
		data = ids_map.get(task_id, {})
		if not data:
			ids_map[task_id] = {}
		elif data.get('ready', False):
			return

		# if not data:
		# 	debug('task not in ids_map', sub='debug.celery', id=task_id)
		# 	return

		# Get remote result
		res = AsyncResult(task_id)
		if not res:
			debug('empty response', sub='celery.data', id=task_id)
			return

		# Set up task state
		data.update({
			'state': res.state,
			'ready': False,
			'results': []
		})

		# Get remote task data
		info = res.info

		# Depending on the task state, info will be either an Exception (FAILURE), a list (SUCCESS), or a dict (RUNNING).
		# - If it's an Exception, it's an unhandled error.
		# - If it's a list, it's the task results.
		# - If it's a dict, it's the custom user metadata.

		if isinstance(info, Exception):
			debug('unhandled exception', obj={'msg': str(info), 'tb': traceback_as_string(info)}, sub='celery.data', id=task_id)
			raise info

		elif isinstance(info, list):
			data['results'] = info
			errors = [e for e in info if e._type == 'error']
			status = 'FAILURE' if errors else 'SUCCESS'
			data['count'] = len([c for c in info if c._source.startswith(data['name'])])
			data['state'] = status

		elif isinstance(info, dict):
			data.update(info)

		# Set ready flag and progress
		ready = data['state'] in ['FAILURE', 'SUCCESS', 'REVOKED']
		data['ready'] = ready
		ids_map[task_id]['ready'] = data['ready']
		if data['ready']:
			data['progress'] = 100
		elif data['results']:
			progresses = [e for e in data['results'] if e._type == 'progress' and e._source == data['full_name']]
			if progresses:
				data['progress'] = progresses[-1].percent

		debug('data', obj=data, sub='celery.data', id=task_id, verbose=True)
		return data

	@staticmethod
	def get_task_ids(result, ids=[]):
		"""Get all Celery task ids recursively.

		Args:
			result (Union[AsyncResult, GroupResult]): Celery result object.
			ids (list): List of ids.
		"""
		if result is None:
			return

		try:
			if isinstance(result, GroupResult):
				CeleryData.get_task_ids(result.parent, ids=ids)

			elif isinstance(result, AsyncResult):
				if result.id not in ids:
					ids.append(result.id)

			if hasattr(result, 'children'):
				children = result.children
				if isinstance(children, list):
					for child in children:
						CeleryData.get_task_ids(child, ids=ids)

			# Browse parent
			if hasattr(result, 'parent') and result.parent:
				CeleryData.get_task_ids(result.parent, ids=ids)

		except kombu.exceptions.DecodeError:
			debug('kombu decode error', sub='celery.data')
			return



import time
from dataclasses import dataclass, field

from secator.output_types import OutputType
from secator.utils import rich_to_ansi


@dataclass
class Progress(OutputType):
	duration: str
	percent: int = 0
	errors: list = field(default_factory=list)
	extra_data: dict = field(default_factory=dict)
	_source: str = field(default='', repr=True)
	_type: str = field(default='progress', repr=True)
	_timestamp: int = field(default_factory=lambda: time.time(), compare=False)
	_uuid: str = field(default='', repr=True, compare=False)
	_context: dict = field(default_factory=dict, repr=True, compare=False)
	_tagged: bool = field(default=False, repr=True, compare=False)
	_duplicate: bool = field(default=False, repr=True, compare=False)
	_related: list = field(default_factory=list, compare=False)

	_table_fields = ['percent', 'duration']
	_sort_by = ('percent',)

	def __post_init__(self):
		super().__post_init__()
		if not 0 <= self.percent <= 100:
			self.percent = 0

	def __str__(self) -> str:
		return f'{self.percent}%'

	def __repr__(self) -> str:
		s = f'[dim]⏳ {self.percent}% ' + '█' * (self.percent // 10) + '[/]'
		if self.errors:
			s += f' [dim red]errors={self.errors}[/]'
		ed = ' '.join([f'{k}={v}' for k, v in self.extra_data.items() if k != 'startedAt' and v])
		s += f' [dim yellow]{ed}[/]'
		return rich_to_ansi(s)



import logging
import time

import pymongo
from bson.objectid import ObjectId
from celery import shared_task

from secator.config import CONFIG
from secator.output_types import FINDING_TYPES
from secator.runners import Scan, Task, Workflow
from secator.utils import debug, escape_mongodb_url

# import gevent.monkey
# gevent.monkey.patch_all()

MONGODB_URL = CONFIG.addons.mongodb.url
MONGODB_UPDATE_FREQUENCY = CONFIG.addons.mongodb.update_frequency
MONGODB_CONNECT_TIMEOUT = CONFIG.addons.mongodb.server_selection_timeout_ms
MONGODB_MAX_POOL_SIZE = CONFIG.addons.mongodb.max_pool_size

logger = logging.getLogger(__name__)

_mongodb_client = None


def get_mongodb_client():
	"""Get or create MongoDB client"""
	global _mongodb_client
	if _mongodb_client is None:
		_mongodb_client = pymongo.MongoClient(
			escape_mongodb_url(MONGODB_URL),
			maxPoolSize=MONGODB_MAX_POOL_SIZE,
			serverSelectionTimeoutMS=MONGODB_CONNECT_TIMEOUT
		)
	return _mongodb_client


def get_runner_dbg(runner):
	"""Runner debug object"""
	return {
		runner.unique_name: runner.status,
		'type': runner.config.type,
		'class': runner.__class__.__name__,
		'caller': runner.config.name,
		**runner.context
	}


def update_runner(self):
	client = get_mongodb_client()
	db = client.main
	type = self.config.type
	collection = f'{type}s'
	update = self.toDict()
	chunk = update.get('chunk')
	_id = self.context.get(f'{type}_chunk_id') if chunk else self.context.get(f'{type}_id')
	debug('to_update', sub='hooks.mongodb', id=_id, obj=get_runner_dbg(self), obj_after=True, obj_breaklines=False, verbose=True)  # noqa: E501
	start_time = time.time()
	if _id:
		db = client.main
		start_time = time.time()
		db[collection].update_one({'_id': ObjectId(_id)}, {'$set': update})
		end_time = time.time()
		elapsed = end_time - start_time
		debug(
			f'[dim gold4]updated in {elapsed:.4f}s[/]', sub='hooks.mongodb', id=_id, obj=get_runner_dbg(self), obj_after=False)  # noqa: E501
		self.last_updated_db = start_time
	else:  # sync update and save result to runner object
		runner = db[collection].insert_one(update)
		_id = str(runner.inserted_id)
		if chunk:
			self.context[f'{type}_chunk_id'] = _id
		else:
			self.context[f'{type}_id'] = _id
		end_time = time.time()
		elapsed = end_time - start_time
		debug(f'in {elapsed:.4f}s', sub='hooks.mongodb', id=_id, obj=get_runner_dbg(self), obj_after=False)


def update_finding(self, item):
	if type(item) not in FINDING_TYPES:
		return item
	start_time = time.time()
	client = get_mongodb_client()
	db = client.main
	update = item.toDict()
	_type = item._type
	_id = ObjectId(item._uuid) if ObjectId.is_valid(item._uuid) else None
	if _id:
		finding = db['findings'].update_one({'_id': _id}, {'$set': update})
		status = 'UPDATED'
	else:
		finding = db['findings'].insert_one(update)
		item._uuid = str(finding.inserted_id)
		status = 'CREATED'
	end_time = time.time()
	elapsed = end_time - start_time
	debug_obj = {
		_type: status,
		'type': 'finding',
		'class': self.__class__.__name__,
		'caller': self.config.name,
		**self.context
	}
	debug(f'in {elapsed:.4f}s', sub='hooks.mongodb', id=str(item._uuid), obj=debug_obj, obj_after=False)  # noqa: E501
	return item


def find_duplicates(self):
	from secator.celery import IN_CELERY_WORKER_PROCESS
	ws_id = self.toDict().get('context', {}).get('workspace_id')
	if not ws_id:
		return
	if not IN_CELERY_WORKER_PROCESS:
		tag_duplicates(ws_id)
	else:
		tag_duplicates.delay(ws_id)


def load_finding(obj):
	finding_type = obj['_type']
	klass = None
	for otype in FINDING_TYPES:
		if finding_type == otype.get_name():
			klass = otype
			item = klass.load(obj)
			item._uuid = str(obj['_id'])
			return item
	return None


def load_findings(objs):
	findings = [load_finding(obj) for obj in objs]
	return [f for f in findings if f is not None]


@shared_task
def tag_duplicates(ws_id: str = None):
	"""Tag duplicates in workspace.

	Args:
		ws_id (str): Workspace id.
	"""
	debug(f'running duplicate check on workspace {ws_id}', sub='hooks.mongodb')
	client = get_mongodb_client()
	db = client.main
	workspace_query = list(
		db.findings.find({'_context.workspace_id': str(ws_id), '_tagged': True}).sort('_timestamp', -1))
	untagged_query = list(
		db.findings.find({'_context.workspace_id': str(ws_id)}).sort('_timestamp', -1))
	# TODO: use this instead when duplicate removal logic is final
	# untagged_query = list(
	# 	db.findings.find({'_context.workspace_id': str(ws_id), '_tagged': False}).sort('_timestamp', -1))
	if not untagged_query:
		debug('no untagged findings. Skipping.', id=ws_id, sub='hooks.mongodb')
		return

	untagged_findings = load_findings(untagged_query)
	workspace_findings = load_findings(workspace_query)
	non_duplicates = []
	duplicates = []
	for item in untagged_findings:
		# If already seen in duplicates
		seen = [f for f in duplicates if f._uuid == item._uuid]
		if seen:
			continue

		# Check for duplicates
		tmp_duplicates = []

		# Check if already present in list of workspace_findings findings, list of duplicates, or untagged_findings
		workspace_dupes = [f for f in workspace_findings if f == item and f._uuid != item._uuid]
		untagged_dupes = [f for f in untagged_findings if f == item and f._uuid != item._uuid]
		seen_dupes = [f for f in duplicates if f == item and f._uuid != item._uuid]
		tmp_duplicates.extend(workspace_dupes)
		tmp_duplicates.extend(untagged_dupes)
		tmp_duplicates.extend(seen_dupes)
		debug(
			f'for item {item._uuid}',
			obj={
				'workspace dupes': len(workspace_dupes),
				'untagged dupes': len(untagged_dupes),
				'seen dupes': len(seen_dupes)
			},
			id=ws_id,
			sub='hooks.mongodb',
			verbose=True)
		tmp_duplicates_ids = list(dict.fromkeys([i._uuid for i in tmp_duplicates]))
		debug(f'duplicate ids: {tmp_duplicates_ids}', id=ws_id, sub='hooks.mongodb', verbose=True)

		# Update latest object as non-duplicate
		if tmp_duplicates:
			duplicates.extend([f for f in tmp_duplicates])
			db.findings.update_one({'_id': ObjectId(item._uuid)}, {'$set': {'_related': tmp_duplicates_ids}})
			debug(f'adding {item._uuid} as non-duplicate', id=ws_id, sub='hooks.mongodb', verbose=True)
			non_duplicates.append(item)
		else:
			debug(f'adding {item._uuid} as non-duplicate', id=ws_id, sub='hooks.mongodb', verbose=True)
			non_duplicates.append(item)

	# debug(f'found {len(duplicates)} total duplicates')

	# Update objects with _tagged and _duplicate fields
	duplicates_ids = list(dict.fromkeys([n._uuid for n in duplicates]))
	non_duplicates_ids = list(dict.fromkeys([n._uuid for n in non_duplicates]))

	search = {'_id': {'$in': [ObjectId(d) for d in duplicates_ids]}}
	update = {'$set': {'_context.workspace_duplicate': True, '_tagged': True}}
	db.findings.update_many(search, update)

	search = {'_id': {'$in': [ObjectId(d) for d in non_duplicates_ids]}}
	update = {'$set': {'_context.workspace_duplicate': False, '_tagged': True}}
	db.findings.update_many(search, update)
	debug(
		'completed duplicates check for workspace.',
		id=ws_id,
		obj={
			'processed': len(untagged_findings),
			'duplicates': len(duplicates_ids),
			'non-duplicates': len(non_duplicates_ids)
		},
		sub='hooks.mongodb')


HOOKS = {
	Scan: {
		'on_init': [update_runner],
		'on_start': [update_runner],
		'on_interval': [update_runner],
		'on_duplicate': [update_finding],
		'on_end': [update_runner],
	},
	Workflow: {
		'on_init': [update_runner],
		'on_start': [update_runner],
		'on_interval': [update_runner],
		'on_duplicate': [update_finding],
		'on_end': [update_runner],
	},
	Task: {
		'on_init': [update_runner],
		'on_start': [update_runner],
		'on_item': [update_finding],
		'on_duplicate': [update_finding],
		'on_interval': [update_runner],
		'on_end': [update_runner]
	}
}
